from __future__ import print_function

__copyright__ = """Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
"""

__doc__ = """ This module provideds a set of functions for working with glyph name
dictionaries. In particular, to load the latest set of recommended Adobe Glyph
names and Unicode values, from one of the scripts in
python/afdko, the set of commands is:
	import agd
	import os
    import fdkutils
	resources_dir = fdkutils.get_resources_dir()
	kAGD_TXTPath = os.path.join(resources_dir, "AGD.txt")
	fp = open(kAGD_TXTPath, "rU")
	agdTextPath = fp.read()
	fp.close()
	gAGDDict = agd.dictionary(agdTextPath)
"""

import re, sys, os, time


#---------------------------------------
def agl():
	"Return a hash of AGL names to Unicode values."
	aglout = {}
	agldata = "A:0041 AE:00C6 AEacute:01FC Aacute:00C1 Abreve:0102 Acircumflex:00C2 Adieresis:00C4 Agrave:00C0 Alpha:0391 Alphatonos:0386 Amacron:0100 Aogonek:0104 Aring:00C5 Aringacute:01FA Atilde:00C3 B:0042 Beta:0392 C:0043 Cacute:0106 Ccaron:010C Ccedilla:00C7 Ccircumflex:0108 Cdotaccent:010A Chi:03A7 D:0044 Dcaron:010E Dcroat:0110 Delta:0394 E:0045 Eacute:00C9 Ebreve:0114 Ecaron:011A Ecircumflex:00CA Edieresis:00CB Edotaccent:0116 Egrave:00C8 Emacron:0112 Eng:014A Eogonek:0118 Epsilon:0395 Epsilontonos:0388 Eta:0397 Etatonos:0389 Eth:00D0 Euro:20AC F:0046 G:0047 Gamma:0393 Gbreve:011E Gcaron:01E6 Gcircumflex:011C Gcommaaccent:0122 Gdotaccent:0120 H:0048 H18533:25CF H18543:25AA H18551:25AB H22073:25A1 Hbar:0126 Hcircumflex:0124 I:0049 IJ:0132 Iacute:00CD Ibreve:012C Icircumflex:00CE Idieresis:00CF Idotaccent:0130 Ifraktur:2111 Igrave:00CC Imacron:012A Iogonek:012E Iota:0399 Iotadieresis:03AA Iotatonos:038A Itilde:0128 J:004A Jcircumflex:0134 K:004B Kappa:039A Kcommaaccent:0136 L:004C Lacute:0139 Lambda:039B Lcaron:013D Lcommaaccent:013B Ldot:013F Lslash:0141 M:004D Mu:039C N:004E Nacute:0143 Ncaron:0147 Ncommaaccent:0145 Ntilde:00D1 Nu:039D O:004F OE:0152 Oacute:00D3 Obreve:014E Ocircumflex:00D4 Odieresis:00D6 Ograve:00D2 Ohorn:01A0 Ohungarumlaut:0150 Omacron:014C Omega:03A9 Omegatonos:038F Omicron:039F Omicrontonos:038C Oslash:00D8 Oslashacute:01FE Otilde:00D5 P:0050 Phi:03A6 Pi:03A0 Psi:03A8 Q:0051 R:0052 Racute:0154 Rcaron:0158 Rcommaaccent:0156 Rfraktur:211C Rho:03A1 S:0053 SF010000:250C SF020000:2514 SF030000:2510 SF040000:2518 SF050000:253C SF060000:252C SF070000:2534 SF080000:251C SF090000:2524 SF100000:2500 SF110000:2502 SF190000:2561 SF200000:2562 SF210000:2556 SF220000:2555 SF230000:2563 SF240000:2551 SF250000:2557 SF260000:255D SF270000:255C SF280000:255B SF360000:255E SF370000:255F SF380000:255A SF390000:2554 SF400000:2569 SF410000:2566 SF420000:2560 SF430000:2550 SF440000:256C SF450000:2567 SF460000:2568 SF470000:2564 SF480000:2565 SF490000:2559 SF500000:2558 SF510000:2552 SF520000:2553 SF530000:256B SF540000:256A Sacute:015A Scaron:0160 Scedilla:015E Scircumflex:015C Scommaaccent:0218 Sigma:03A3 T:0054 Tau:03A4 Tbar:0166 Tcaron:0164 Tcommaaccent:0162 Theta:0398 Thorn:00DE U:0055 Uacute:00DA Ubreve:016C Ucircumflex:00DB Udieresis:00DC Ugrave:00D9 Uhorn:01AF Uhungarumlaut:0170 Umacron:016A Uogonek:0172 Upsilon:03A5 Upsilon1:03D2 Upsilondieresis:03AB Upsilontonos:038E Uring:016E Utilde:0168 V:0056 W:0057 Wacute:1E82 Wcircumflex:0174 Wdieresis:1E84 Wgrave:1E80 X:0058 Xi:039E Y:0059 Yacute:00DD Ycircumflex:0176 Ydieresis:0178 Ygrave:1EF2 Z:005A Zacute:0179 Zcaron:017D Zdotaccent:017B Zeta:0396 a:0061 aacute:00E1 abreve:0103 acircumflex:00E2 acute:00B4 acutecomb:0301 adieresis:00E4 ae:00E6 aeacute:01FD afii00208:2015 afii10017:0410 afii10018:0411 afii10019:0412 afii10020:0413 afii10021:0414 afii10022:0415 afii10023:0401 afii10024:0416 afii10025:0417 afii10026:0418 afii10027:0419 afii10028:041A afii10029:041B afii10030:041C afii10031:041D afii10032:041E afii10033:041F afii10034:0420 afii10035:0421 afii10036:0422 afii10037:0423 afii10038:0424 afii10039:0425 afii10040:0426 afii10041:0427 afii10042:0428 afii10043:0429 afii10044:042A afii10045:042B afii10046:042C afii10047:042D afii10048:042E afii10049:042F afii10050:0490 afii10051:0402 afii10052:0403 afii10053:0404 afii10054:0405 afii10055:0406 afii10056:0407 afii10057:0408 afii10058:0409 afii10059:040A afii10060:040B afii10061:040C afii10062:040E afii10065:0430 afii10066:0431 afii10067:0432 afii10068:0433 afii10069:0434 afii10070:0435 afii10071:0451 afii10072:0436 afii10073:0437 afii10074:0438 afii10075:0439 afii10076:043A afii10077:043B afii10078:043C afii10079:043D afii10080:043E afii10081:043F afii10082:0440 afii10083:0441 afii10084:0442 afii10085:0443 afii10086:0444 afii10087:0445 afii10088:0446 afii10089:0447 afii10090:0448 afii10091:0449 afii10092:044A afii10093:044B afii10094:044C afii10095:044D afii10096:044E afii10097:044F afii10098:0491 afii10099:0452 afii10100:0453 afii10101:0454 afii10102:0455 afii10103:0456 afii10104:0457 afii10105:0458 afii10106:0459 afii10107:045A afii10108:045B afii10109:045C afii10110:045E afii10145:040F afii10146:0462 afii10147:0472 afii10148:0474 afii10193:045F afii10194:0463 afii10195:0473 afii10196:0475 afii10846:04D9 afii299:200E afii300:200F afii301:200D afii57381:066A afii57388:060C afii57392:0660 afii57393:0661 afii57394:0662 afii57395:0663 afii57396:0664 afii57397:0665 afii57398:0666 afii57399:0667 afii57400:0668 afii57401:0669 afii57403:061B afii57407:061F afii57409:0621 afii57410:0622 afii57411:0623 afii57412:0624 afii57413:0625 afii57414:0626 afii57415:0627 afii57416:0628 afii57417:0629 afii57418:062A afii57419:062B afii57420:062C afii57421:062D afii57422:062E afii57423:062F afii57424:0630 afii57425:0631 afii57426:0632 afii57427:0633 afii57428:0634 afii57429:0635 afii57430:0636 afii57431:0637 afii57432:0638 afii57433:0639 afii57434:063A afii57440:0640 afii57441:0641 afii57442:0642 afii57443:0643 afii57444:0644 afii57445:0645 afii57446:0646 afii57448:0648 afii57449:0649 afii57450:064A afii57451:064B afii57452:064C afii57453:064D afii57454:064E afii57455:064F afii57456:0650 afii57457:0651 afii57458:0652 afii57470:0647 afii57505:06A4 afii57506:067E afii57507:0686 afii57508:0698 afii57509:06AF afii57511:0679 afii57512:0688 afii57513:0691 afii57514:06BA afii57519:06D2 afii57534:06D5 afii57636:20AA afii57645:05BE afii57658:05C3 afii57664:05D0 afii57665:05D1 afii57666:05D2 afii57667:05D3 afii57668:05D4 afii57669:05D5 afii57670:05D6 afii57671:05D7 afii57672:05D8 afii57673:05D9 afii57674:05DA afii57675:05DB afii57676:05DC afii57677:05DD afii57678:05DE afii57679:05DF afii57680:05E0 afii57681:05E1 afii57682:05E2 afii57683:05E3 afii57684:05E4 afii57685:05E5 afii57686:05E6 afii57687:05E7 afii57688:05E8 afii57689:05E9 afii57690:05EA afii57716:05F0 afii57717:05F1 afii57718:05F2 afii57793:05B4 afii57794:05B5 afii57795:05B6 afii57796:05BB afii57797:05B8 afii57798:05B7 afii57799:05B0 afii57800:05B2 afii57801:05B1 afii57802:05B3 afii57803:05C2 afii57804:05C1 afii57806:05B9 afii57807:05BC afii57839:05BD afii57841:05BF afii57842:05C0 afii57929:02BC afii61248:2105 afii61289:2113 afii61352:2116 afii61573:202C afii61574:202D afii61575:202E afii61664:200C afii63167:066D afii64937:02BD agrave:00E0 aleph:2135 alpha:03B1 alphatonos:03AC amacron:0101 ampersand:0026 angle:2220 angleleft:2329 angleright:232A anoteleia:0387 aogonek:0105 approxequal:2248 aring:00E5 aringacute:01FB arrowboth:2194 arrowdblboth:21D4 arrowdbldown:21D3 arrowdblleft:21D0 arrowdblright:21D2 arrowdblup:21D1 arrowdown:2193 arrowleft:2190 arrowright:2192 arrowup:2191 arrowupdn:2195 arrowupdnbse:21A8 asciicircum:005E asciitilde:007E asterisk:002A asteriskmath:2217 at:0040 atilde:00E3 b:0062 backslash:005C bar:007C beta:03B2 block:2588 braceleft:007B braceright:007D bracketleft:005B bracketright:005D breve:02D8 brokenbar:00A6 bullet:2022 c:0063 cacute:0107 caron:02C7 carriagereturn:21B5 ccaron:010D ccedilla:00E7 ccircumflex:0109 cdotaccent:010B cedilla:00B8 cent:00A2 chi:03C7 circle:25CB circlemultiply:2297 circleplus:2295 circumflex:02C6 club:2663 colon:003A colonmonetary:20A1 comma:002C congruent:2245 copyright:00A9 currency:00A4 d:0064 dagger:2020 daggerdbl:2021 dcaron:010F dcroat:0111 degree:00B0 delta:03B4 diamond:2666 dieresis:00A8 dieresistonos:0385 divide:00F7 dkshade:2593 dnblock:2584 dollar:0024 dong:20AB dotaccent:02D9 dotbelowcomb:0323 dotlessi:0131 dotmath:22C5 e:0065 eacute:00E9 ebreve:0115 ecaron:011B ecircumflex:00EA edieresis:00EB edotaccent:0117 egrave:00E8 eight:0038 element:2208 ellipsis:2026 emacron:0113 emdash:2014 emptyset:2205 endash:2013 eng:014B eogonek:0119 epsilon:03B5 epsilontonos:03AD equal:003D equivalence:2261 estimated:212E eta:03B7 etatonos:03AE eth:00F0 exclam:0021 exclamdbl:203C exclamdown:00A1 existential:2203 f:0066 female:2640 figuredash:2012 filledbox:25A0 filledrect:25AC five:0035 fiveeighths:215D florin:0192 four:0034 fraction:2044 franc:20A3 g:0067 gamma:03B3 gbreve:011F gcaron:01E7 gcircumflex:011D gcommaaccent:0123 gdotaccent:0121 germandbls:00DF gradient:2207 grave:0060 gravecomb:0300 greater:003E greaterequal:2265 guillemotleft:00AB guillemotright:00BB guilsinglleft:2039 guilsinglright:203A h:0068 hbar:0127 hcircumflex:0125 heart:2665 hookabovecomb:0309 house:2302 hungarumlaut:02DD hyphen:002D i:0069 iacute:00ED ibreve:012D icircumflex:00EE idieresis:00EF igrave:00EC ij:0133 imacron:012B infinity:221E integral:222B integralbt:2321 integraltp:2320 intersection:2229 invbullet:25D8 invcircle:25D9 invsmileface:263B iogonek:012F iota:03B9 iotadieresis:03CA iotadieresistonos:0390 iotatonos:03AF itilde:0129 j:006A jcircumflex:0135 k:006B kappa:03BA kcommaaccent:0137 kgreenlandic:0138 l:006C lacute:013A lambda:03BB lcaron:013E lcommaaccent:013C ldot:0140 less:003C lessequal:2264 lfblock:258C lira:20A4 logicaland:2227 logicalnot:00AC logicalor:2228 longs:017F lozenge:25CA lslash:0142 ltshade:2591 m:006D macron:00AF male:2642 minus:2212 minute:2032 mu:03BC multiply:00D7 musicalnote:266A musicalnotedbl:266B n:006E nacute:0144 napostrophe:0149 ncaron:0148 ncommaaccent:0146 nine:0039 notelement:2209 notequal:2260 notsubset:2284 ntilde:00F1 nu:03BD numbersign:0023 o:006F oacute:00F3 obreve:014F ocircumflex:00F4 odieresis:00F6 oe:0153 ogonek:02DB ograve:00F2 ohorn:01A1 ohungarumlaut:0151 omacron:014D omega:03C9 omega1:03D6 omegatonos:03CE omicron:03BF omicrontonos:03CC one:0031 onedotenleader:2024 oneeighth:215B onehalf:00BD onequarter:00BC onethird:2153 openbullet:25E6 ordfeminine:00AA ordmasculine:00BA orthogonal:221F oslash:00F8 oslashacute:01FF otilde:00F5 p:0070 paragraph:00B6 parenleft:0028 parenright:0029 partialdiff:2202 percent:0025 period:002E periodcentered:00B7 perpendicular:22A5 perthousand:2030 peseta:20A7 phi:03C6 phi1:03D5 pi:03C0 plus:002B plusminus:00B1 prescription:211E product:220F propersubset:2282 propersuperset:2283 proportional:221D psi:03C8 q:0071 question:003F questiondown:00BF quotedbl:0022 quotedblbase:201E quotedblleft:201C quotedblright:201D quoteleft:2018 quotereversed:201B quoteright:2019 quotesinglbase:201A quotesingle:0027 r:0072 racute:0155 radical:221A rcaron:0159 rcommaaccent:0157 reflexsubset:2286 reflexsuperset:2287 registered:00AE revlogicalnot:2310 rho:03C1 ring:02DA rtblock:2590 s:0073 sacute:015B scaron:0161 scedilla:015F scircumflex:015D scommaaccent:0219 second:2033 section:00A7 semicolon:003B seven:0037 seveneighths:215E shade:2592 sigma:03C3 sigma1:03C2 similar:223C six:0036 slash:002F smileface:263A space:0020 spade:2660 sterling:00A3 suchthat:220B summation:2211 sun:263C t:0074 tau:03C4 tbar:0167 tcaron:0165 tcommaaccent:0163 therefore:2234 theta:03B8 theta1:03D1 thorn:00FE three:0033 threeeighths:215C threequarters:00BE tilde:02DC tildecomb:0303 tonos:0384 trademark:2122 triagdn:25BC triaglf:25C4 triagrt:25BA triagup:25B2 two:0032 twodotenleader:2025 twothirds:2154 u:0075 uacute:00FA ubreve:016D ucircumflex:00FB udieresis:00FC ugrave:00F9 uhorn:01B0 uhungarumlaut:0171 umacron:016B underscore:005F underscoredbl:2017 union:222A universal:2200 uogonek:0173 upblock:2580 upsilon:03C5 upsilondieresis:03CB upsilondieresistonos:03B0 upsilontonos:03CD uring:016F utilde:0169 v:0076 w:0077 wacute:1E83 wcircumflex:0175 wdieresis:1E85 weierstrass:2118 wgrave:1E81 x:0078 xi:03BE y:0079 yacute:00FD ycircumflex:0177 ydieresis:00FF yen:00A5 ygrave:1EF3 z:007A zacute:017A zcaron:017E zdotaccent:017C zero:0030 zeta:03B6 "
	for n in agldata.split():
		nn = n.split(':')
		if len(nn) == 2: aglout[nn[0]] = nn[1]
	return aglout

aglist = agl() # Because 'agl()' is slow, 'aglist' is defined here for use in this module



#===========================================================
# A glyph entry from a Glyph Dictionary

re_entry = re.compile(r'(?m)^[\t ]+(\S+?): +(\S.*?)$') # regexp: data entry line
re_name = re.compile(r'[A-Za-z0-9_\.]+') # regexp: glyph name
class glyph:
	def __init__(self, name):
		self.name = name
		self.uni = None # The Unicode
		self.fin = None # The final name
		self.ali = [] # list of name aliases
		self.sub = [] # list of substitutions
		self.set = [] # list of glyph set IDs
		self.min = None # minuscule
		self.maj = None # majuscule
		self.cmp = None # components used to create this composite glyph
		self.other = {} # Hash of any unknown tags

	# add entry data from input text:
	def parse(self, intext):
		ee = re_entry.findall(intext) # get all data entries
		if len(ee) < 1: return False # return if no data entries
		for e in ee:
			k, x = e # get key and its data
			if k == 'uni':
				self.uni = re.compile(r'[0-9A-F]{4}').search(x).group() # Unicode value
			elif k == 'fin': self.fin = re_name.search(x).group() # final name
			elif k == 'ali':
				for a in re_name.findall(x): self.ali.append(a) # name aliases
			elif k == 'sub': self.sub = x.split() # substitutions
			elif k == 'set': self.set.extend(x.split()) # glyph sets
			elif k == 'min': self.min = re_name.search(x).group() # minuscule
			elif k == 'maj': self.maj = re_name.search(x).group() # majuscule
			elif k == 'cmp': self.cmp = x # components
			else: self.other[k] = x # any unknown entry key
		self.check() # re-check glyph data
		return True

	# Check a glyph for internal conflicts, problems, etc.
	def check(self):
		m = [] # holds reports of what was fixed
		u = re.compile(r'uni([0-9A-F]{4})$') # Unicode value regexp
		n = {}
		# process aliases: remove duplicates, uninames, final names:
		for a in self.ali:
			# Remove alias if it matches the index name:
			if a == self.name:
				m.append("Removed alias '%s' from glyph '%s' (same as index name)." % (a, self.name))
				continue
			# Remove alias if it matches the final name:
			elif self.fin and a == self.fin:
				m.append("Removed alias '%s' from glyph '%s' (same as final name)." % (a, self.name))
				continue
			# Remove alias if it is a uni-name:
			elif u.match(a):
				m.append("Removed alias '%s' from glyph '%s' (uni-name for Unicode value)." % (a, self.name))
				continue
			else: n[a] = 1 # pass the alias
		self.ali = sorted(n.keys()) # sorted list of passed aliases
		# check final name:
		if self.fin and self.fin == self.name:
			m.append("Removed final name '%s' from glyph '%s' (same as index name)." % (self.fin, self.name))
			self.fin = None
		elif self.fin and self.uni and u.match(self.fin) and u.match(self.fin).group(1) != self.uni:
			m.append("Removed final name '%s' from glyph '%s'. Conflicts with Unicode %s." % (self.fin, self.name, self.uni))
			self.fin = None
		# Look for Unicode values in glyph names:
		if not self.uni:
			if self.fin and u.match(self.fin):
				m.append("Found Unicode value in final name '%s' of glyph '%s'." % (self.fin, self.name))
				self.uni = u.match(self.fin).group(1) # Get Unicode from final name
			else:
				for a in self.ali:
					if u.match(a):
						m.append("Found Unicode value in alias '%s' of glyph '%s'." % (a, self.name))
						self.uni = u.match(a).group(1)
						break
		return m

	# Return a list of characters mapped from a glyph name:
	def characters(self):
		c = self.name.split('.', 1)
		return c[0].split('_')

	# Return a list of all glyph names/aliases, including uni-names and final names
	def aliases(self, addglyphs=[]):
		self.check()
		n = {}
		n[self.name] = 1
		if self.fin: n[self.fin] = 1
		if self.uni: n[self.uniname()] = 1
		for i in self.ali: n[i] = 1 # add the glyph's own alias names
		for i in addglyphs: n[i] = 1 # add any additional alias names
		nn = sorted(n.keys())
		return nn

	# Return a text entry
	def entry(self):
		e = [] # Output lines
		e.append(self.name) # index name
		if self.uni: e.append("\tuni: %s" % self.uni) # Unicode value
		if self.fin: e.append("\tfin: %s" % self.fin) # final name
		if self.ali: e.append("\tali: %s" % ' '.join(self.ali)) # name aliases
		if self.sub: e.append("\tsub: %s" % ' '.join(self.sub)) # substitutions
		if self.set: e.append("\tset: %s" % ' '.join(self.set)) # glyph sets
		if self.min: e.append("\tmin: %s" % self.min) # minuscule
		if self.maj: e.append("\tmaj: %s" % self.maj) # majuscule
		if self.cmp: e.append("\tcmp: %s" % self.cmp) # components
		if self.other:
			otherkeys = sorted(self.other.keys())
			for k in otherkeys: e.append("\t%s: %s" % (k, self.other[k])) # other keys
		return '\n'.join(e) + '\n'

	# return a uni-name for the glyph's Unicode value:
	def uniname(self):
		if self.uni: return "uni%s"%self.uni
		else: return None

	# return the glyph's final name, which is a specified final name, or the work name:
	def finalname(self):
		return (self.fin or self.name or None)

	# return a glyph alias entry, with optionally-specified work name:
	def aliasline(self, workname=None):
		lineout = [] # output line
		lineout.append(self.fin or self.name)
		lineout.append(workname or self.name)
		if self.uni: lineout.append(self.uniname())
		return "\t".join(lineout)

	# report whether the glyph has a Unicode value in the PUA range (E000-F8FF):
	def haspua(self):
		if not self.uni: return False
		u = int(self.uni, 16)
		if 57343 < u < 63744: return True
		else: return False


#---------------------------------------
# A Glyph Dictionary

class dictionary:
	def __init__(self, intext=None):
		self.list = [] # Ordered list of glyph names
		self.glyphs = {} # Dictionary of glyph names and parents
		self.index = {} # Hash of all glyph/alias names
		self.unicode = {} # Hash of Unicode values
		self.messages = [] # A list of error/warning messages
		if intext: self.parse(intext)

	# Read text, convert to glyph entries, and add to dictionary:
	def parse(self, intext, priority=1):
		intext += '\n' # ensure the file ends with a newline for grep purposes
		re_entry = re.compile(r'(?m)^([A-Za-z0-9\._]+)\n((?:[\t ]+\S.*?\n)*)') # regexp for glyph entry text
		ee = re_entry.findall(intext) # find all glyph entries
		if len(ee) < 1: return False # return if no entries found
		for e in ee:
			gname, ginfo = e
			g = glyph(gname) # create new glyph object
			g.parse(ginfo)
			self.add(g, priority)
		return True

	# Find a glyph via index lookup, or by Unicode value:
	def glyph(self, n):
		if not n: return None
		elif n in self.index: return self.glyphs[self.index[n]] # lookup by glyph name
		elif n in self.unicode: return self.glyphs[self.unicode[n]] # lookup by Unicode value
		else: return None

	# Remove an existing glyph entry
	def remove(self, n):
		if n in self.glyphs:
			g = self.glyphs[n]
			for a in g.aliases():
				if a in self.index: del self.index[a] # remove all index entries
			if g.uni and g.uni in self.unicode: del self.unicode[g.uni] # remove unicode entry
			del self.glyphs[n] # delete the glyph entry
			i = self.list.index(n) # get order position of glyph
			del self.list[i] # remove glyph name from order list
			return i
		else: return None

	# Find and remove an alias:
	def removealias(self, n):
		if n in self.glyphs:
			self.messages.append("Cannot remove alias '%s'. (Name exists as index name.)" % n)
			return False
		elif n in self.index:
			g = self.glyphs[self.index[n]] # the glyph entry containing the alias
			if g.uni and g.uniname() == n:
				self.messages.append("Cannot remove alias '%s' from glyph '%s'. (Alias is derived from Unicode value.)" % (n, g.name))
				return False
			if g.fin and g.fin == n:
				self.messages.append("Removed final name '%s' from glyph '%s'." % (n, g.name))
				g.fin = None
			else:
				self.messages.append("Removed alias '%s' from glyph '%s'." % (n, g.name))
				g.ali.remove(n)
			del self.index[n]
			g.check()
			return True
		else:
			self.messages.append("Cannot remove alias '%s'. (Name not found.)" % n)
			return False

	# Find and remove a Unicode value:
	def removeunicode(self, u):
		if u in self.unicode:
			if self.unicode[u] in self.glyphs:
				g = self.glyphs[self.unicode[u]]
				ua = g.uniname() # uni-name for Unicode value
				if ua in self.index: del self.index[ua]
				if ua in g.ali: g.ali.remove(ua) # remove uni-name from aliases
				g.uni = None # remove Unicode value
			del self.unicode[u]
			return True
		return False

	# Add a glyph object to the dictionary and perform cross-checking;
	# merge incoming data according to specified priority:
	#  priority=3: New entries completely replace existing entries
	#  priority=2: New data merged, override existing data
	#  priority=1: New data merged, defer to existing data
	#  priority=0: New entries defer entirely to existing entries
	def add(self, g, priority=1):
		m = self.messages
		u = re.compile(r'uni([0-9A-F]{4})$') # Unicode value regexp
		addplace = None # position of added glyph in glyph order
		g.check() # check glyph before adding

		# (t) The existing glyph, if any
		t = self.glyph(g.name) or self.glyph(g.uni) # the target glyph which matches the incoming glyph (g) (or None)
		for a in g.aliases():
			if t: break
			t = self.glyph(a)

		# if a target glyph (t) is found:
		if t:
			# Stop and return false if Priority is 0:
			if priority == 0:
				m.append("Cannot add glyph '%s'. Conflicts with existing glyph '%s' (priority 0)." % (g.name, t.name))
				return False

			# If Priority is 3, do not merge any data from (t) into (g)
			elif priority == 3:
				m.append("Replaced existing entry '%s' with new entry '%s' (priority 3)." % (t.name, g.name))

			# If Priority is 1 or 2, merge data from (t) into (g):
			else:
				m.append("Merged new glyph '%s' with existing glyph '%s' (priority %s)." % (g.name, t.name, priority))
				g.ali = t.aliases(g.aliases()) # append new aliases
				for i in t.set:
					if not i in g.set: g.set.append(i) # append new glyph sets
				for i in t.sub:
					if not i in g.sub: g.sub.append(i) # append new features

				if priority == 2:
					if t.uni and not g.uni: g.uni = t.uni # Unicode value
					if t.fin and not g.fin: g.fin = t.fin # final name
					if t.min and not g.min: g.min = t.min # minuscule
					if t.maj and not g.maj: g.maj = t.maj # majuscule
					if t.cmp and not g.cmp: g.cmp = t.cmp # components
					for i in t.other:
						if not i in g.other: g.other[i] = t.other[i]

				else:
					g.name = t.name
					if t.uni: g.uni = t.uni # Unicode value
					if g.fin and not g.fin in g.ali: g.ali.append(g.fin) # copy final name to aliases
					if t.fin: g.fin = t.fin # final name
					if t.min: g.min = t.min # minuscule
					if t.maj: g.maj = t.maj # majuscule
					if t.cmp: g.cmp = t.cmp # majuscule
					for i in t.other: g.other[i] = t.other[i]
			addplace = self.remove(t.name) # remove the conflicting glyph

		else: m.append("Adding new glyph: %s" % g.name)

		# check glyph data against dictionary:
		for n in g.ali:
			# check for existing alias names:
			if n in self.index:
				if priority > 1: self.removealias(n) # remove the alias from the dictionary
				else: g.ali.remove(n) # remove the new alias
		if g.uni and g.uni in self.unicode:
			# check for existing Unicode value
			if priority > 1: self.removeunicode(g.uni)
			else:
				if g.fin and u.match(g.fin) and u.match(g.fin).group(1) == g.uni: g.fin = None # remove incoming final uni-name
				g.uni = None # remove Unicode value
		if g.fin and g.fin in self.index:
			if priority > 1: self.glyphs[self.index[g.fin]].fin = None
			else: g.fin = None
#		m.extend(g.check()) # re-check the new glyph

		# Add the glyph to the dictionary:
		if addplace is None: self.list.append(g.name)
		else: self.list.insert(addplace, g.name) # replace the old glyph with the new glyph in the order list
		self.glyphs[g.name] = g
		for a in g.aliases(): self.index[a] = g.name # add aliases to index (including self, uni-name and final name)
		if g.uni: self.unicode[g.uni] = g.name  # add the Unicode value to the Unicode index
#		m.append("Added glyph '%s'." % g.name)

		return True

	# Create AGD text output of all entries:
	def entries(self):
		ee = []
		for g in self.list: ee.append(self.glyphs[g].entry())
		return ''.join(ee)

	# Create a list of Unicodes in the dictionary, sorted by Unicode value:
	def unicodes(self):
		x = []
		uu = sorted(self.unicode.keys())
		print("Unicodes: %s" % len(uu))
		for u in uu: x.append("%s\t%s" % (self.glyphs[self.unicode[u]].uni, self.unicode[u]))
		return "\n".join(x)

	# Return all logged messages:
	def report(self):
		x = [] # collect output lines
		x.append("\n== Glyph Dictionary messages ===================\n") # output header
		if len(self.messages) < 1: x.append("[No messages]") # if no messages
		else:
			for i in self.messages: x.append(i)
		x.append("\n")
		return "\n".join(x)

	# Return a GlyphOrderAndAliasDB file:
	# Subset and sort output with glyphs list nn
	def aliasfile(self, nn=None):
		textout = []
		if nn:
			outlist = []
			for n in nn:
				if n in self.index: outlist.append(n)
		else: outlist = self.list
		for workname in outlist:
			g = self.glyph(workname)
			textout.append(g.aliasline(workname))
		return '\n'.join(textout)

	# Change the order of the glyphs in the dictionary
	# Sort to match a list of glyph names (s)
	# If no order list s submitted, order is alphabetical
	def sort(self, s):
		nmatch = {} # the number of matches to the order list
		neworder = [] # the new sort order
		for n in s:
			if n in self.glyphs:
				neworder.append(n)
				nmatch[n] = 1
		for n in self.list:
			if not n in nmatch: neworder.append(n)
		self.list = neworder # set the new glyph order
		return len(nmatch) # return the number of name matches in the sort list







#---------------------------------------
# Feature file functions:
#---------------------------------------
# An OpenType glyph substitution object (s)
# s.input: a list of input glyphs to the substitution
# s.output: a list of output glyphs to the substitution
# s.type: the substitution's lookupType:
#   Lookup type 1: single substitution
#   Lookup type 3: alternate substitution
#   Lookup type 4: ligature substitution
#   Lookup type 6: chaining contextual substitution
class substitution:
	def __init__(self, a=[], b=[], intype=0):
		if isinstance(a, str): self.input = [a] # if a is string, not list
		else: self.input = list(a) # List of input glyphs
		if isinstance(b, str): self.output = [b] # if b is string, not list
		else: self.output = list(b) # List of output glyphs

		if len(self.input) == 1 and len(self.output) == 1: self.type = 1
		elif len(self.input) > 1 and len(self.output) == 1: self.type = 4
		elif len(self.input) == 1 and len(self.output) > 1: self.type = 3
		else: self.type = 0 # The GSUB lookup type

		if intype: self.type = intype # Set input override type

	# return a single glyph substitution as a tuple if the input and
	# output both have one glyphs, otherwise return None:
	def pair(self):
		if len(self.input) == 1 and len(self.output) == 1:
			return (self.input[0], self.output[0])
		else: return None

	# return a feature file substitution statement:
	def feature(self):
		len_a = len(self.input)
		len_b = len(self.output)
		subword = 'by'
		if len_a > 0:
			a = ' '.join(self.input)
			if len_a > 1 and self.type != 4: a = '[%s]' % a
		else: a = ''

		if len_b > 1: b = '[%s]' % ' '.join(self.output)
		elif len_b == 1: b = self.output[0]
		else: b = ''

		if len_a == 1 and len_b > 1: subword = 'from'

		if a and b: return '\tsub %s %s %s;' % (a, subword, b)
		elif b and not a: return '\t# %s' % b
		else: return ''

#-----------------------------------------------------------
# Looks through a list of substitution instances (ss) and
# consolidates one-for-one entries into many-for-one or one-from-many
def consolidate(ss):
	glyphs_a = {} # collects matching input glyphs
	glyphs_b = {} # collects matching output glyphs
	log_a = {} # used for logging input glyphs
	log_b = {} # used for logging output glyphs
	newlist = [] # holds consolidated substitution objects


	classidx = {} # dictionary of class glyphs and their class index
	classpairs = [] # holds substitution class pairs
	classpass = [] # holds substitutions passed by the class-checker
	glyphclasses = [
		"zero one two three four five six seven eight nine",
		"A B C D E F G H I J K L M N O P Q R S T U V W X Y Z",
		"a b c d e f g h i j k l m n o p q r s t u v w x y z"
	]

	# index all glyph names in glyph classes:
	for c in range(len(glyphclasses)):
		classpairs.append([])
		for n in glyphclasses[c].split(): classidx[n] = c

	# look through substitutions for glyph which match classes:
	for s in ss:
		if not s.pair():
			classpass.append(s) # pass subs which are not 1:1
			continue
		a, b = s.pair()
		if a in classidx:
			i = classidx[a] # the index of the matching class
			classpairs[i].append(s)
		elif b in classidx:
			i = classidx[b] # the index of the matching class
			classpairs[i].append(s)
		else: classpass.append(s) # pass subs which do not match

	# collate substitutions which match:
	for c in range(len(classpairs)):
		if not classpairs[c]: continue
		class_a = []
		class_b = []
		for s in classpairs[c]:
			a, b = s.pair()
			class_a.append(a)
			class_b.append(b)
		newlist.append(substitution(class_a, class_b, 1)) # add a new substitution object


	# consolidate matching input and output glyphs
	# only in one-to-one substitutions:
	for s in classpass:
		if not s.pair(): continue
		a, b = s.pair()
		if not a in glyphs_a: glyphs_a[a] = []
		glyphs_a[a].append(b)
		if not b in glyphs_b: glyphs_b[b] = []
		glyphs_b[b].append(a)

	# process consolidated glyphs, create new multi-glyph substitution objects:
	for s in classpass:
		if s.pair():
			a, b = s.pair()
			# if many input glyphs (many-to-one):
			if len(glyphs_b[b]) > 1:
				if not b in log_b:
					newlist.append(substitution(glyphs_b[b], b, 1)) # create a new substitution, lookup type 1
					log_b[b] = 1
			# if many output glyphs (one-from-many):
			elif len(glyphs_a[a]) > 1:
				if not a in log_a:
					newlist.append(substitution(a, glyphs_a[a], 3)) # create a new substitution, lookup type 3
					log_a[a] = 1
			else: newlist.append(s)
		else: newlist.append(s)
	return newlist

#---------------------------------------
# Create feature file code from a dictionary (d) and
# optional list of glyph names (nn)
def makefeatures(d, nn=None):
	outlist = [] # List to collect output lines
	features = {} # Hash of feature names
	glyphnames = {} # Hash to hold real names matched to list names
	glyphpass = []

	nn = nn or d.list # Get list of glyphs from dictionary if not submitted

	# check list of glyph names:
	for n in nn:
		g = d.glyph(n)
		if g:
			glyphnames[g.name] = n
			glyphpass.append(g)

	for g in glyphpass:
		subpass = []
		if not g.sub: continue
		# Parse substitution data and create substitution objects:
		for subst in g.sub:
			ss = subst.split('+')
			f = ss.pop() # get the feature tag

			# filter substitution input glyphs which do not exist:
			if ss:
				for s in ss:
					if s in glyphnames: subpass.append(glyphnames[s])
				if not subpass: continue # skip this substitution if all input glyphs were not found

			if not (f in features): features[f] = [] # create a feature tag entry
			features[f].append(substitution(subpass, glyphnames[g.name])) # Append a substitution instance to the feature's list

	flist = sorted(features.keys()) # all feature tags

	# Create feature text output:
	for f in flist:
		outlist.append('feature %s {' % f)
		cs = consolidate(features[f]) # consolidate the feature's substitutions
		for s in cs: outlist.append(s.feature())
		outlist.append('} %s;\n' % f)

	return '\n'.join(outlist)







#-----------------------------------------------------------
# Make a derivedchars text entry from a glyph
def makederived(g):
	dvout = [] # output lines
	if not g.cmp: return "" # return if the glyph has no component entry
	dvout.append("/%s DerivedName" % g.name)
	dvbase = False
	for ca in g.cmp.split():
		dvadd = [] # the composite's components
		for cb in ca.split('+'):
			if len(dvadd) > 0: dvop = "AddCentered"
			elif not dvbase:
				dvop = "Base"
				dvbase = True
			else: dvop = "AddAdjacent"
			dvadd.append("(%s) %s" % (cb, dvop))
		dvout.extend(dvadd)
	dvout.append("EndDerived")
	return "\n".join(dvout)


#---------------------------------------
# Create derivedchars entries from composite information in
# dictionary (d). Produce entries for (glyphlist) or all glyphs
# in (d) if (glyphlist) is None.
# A composite entry will be skipped if all its components
# are not in (glyphlist).
def derivedchars(d, glyphlist=None):
	cmpsplit = re.compile(r' |\+') # regexp for splitting 'cmp' text
	dvout = [] # derivedchars output
	glyphlist = glyphlist or d.list
	for n in glyphlist:
		g = d.glyph(n)
		if not g: continue
		cflag = False # flag to check for missing components
		if g.cmp:
			nn = cmpsplit.split(g.cmp)
			for n in nn:
				if not n in glyphlist: cflag = True
			if cflag: continue
			dvout.append(makederived(g))
	return "\n\n".join(dvout)





#-----------------------------------------------------------
# Parse a GlyphOrderAndAliasDB
# return a list of glyph objects parsed from GlyphOrderAndAliasDB file (goadb)
def parsealiasfile(goadb):
	gg = [] # list of output glyphs
	ga_entry = re.compile(r'(?m)^([A-Za-z0-9\._]+)\t([A-Za-z0-9\._]+)(?:\t(?:uni([0-9A-F]{4}))?)?$')
	ee = ga_entry.findall(goadb) # find all glyph alias entries
	if len(ee) < 1: return gg # return if no entries found
	for e in ee:
		g = glyph(e[1]) # create new glyph object
		if e[0] != e[1]: g.fin = e[0]
		if e[2]: g.uni = e[2]
		gg.append(g)
	return gg



#---------------------------------------
# CFF glyph order
# return a list of glyph names from the CFF specification
# with a submitted list of glyphs (nn), returns that list sorted, with extra glyphs at the end
def cfforder(nn=None):
	c = ".notdef space exclam quotedbl numbersign dollar percent ampersand quoteright parenleft parenright asterisk plus comma hyphen period slash zero one two three four five six seven eight nine colon semicolon less equal greater question at A B C D E F G H I J K L M N O P Q R S T U V W X Y Z bracketleft backslash bracketright asciicircum underscore quoteleft a b c d e f g h i j k l m n o p q r s t u v w x y z braceleft bar braceright asciitilde exclamdown cent sterling fraction yen florin section currency quotesingle quotedblleft guillemotleft guilsinglleft guilsinglright fi fl endash dagger daggerdbl periodcentered paragraph bullet quotesinglbase quotedblbase quotedblright guillemotright ellipsis perthousand questiondown grave acute circumflex tilde macron breve dotaccent dieresis ring cedilla hungarumlaut ogonek caron emdash AE ordfeminine Lslash Oslash OE ordmasculine ae dotlessi lslash oslash oe germandbls onesuperior logicalnot mu trademark Eth onehalf plusminus Thorn onequarter divide brokenbar degree thorn threequarters twosuperior registered minus eth multiply threesuperior copyright Aacute Acircumflex Adieresis Agrave Aring Atilde Ccedilla Eacute Ecircumflex Edieresis Egrave Iacute Icircumflex Idieresis Igrave Ntilde Oacute Ocircumflex Odieresis Ograve Otilde Scaron Uacute Ucircumflex Udieresis Ugrave Yacute Ydieresis Zcaron aacute acircumflex adieresis agrave aring atilde ccedilla eacute ecircumflex edieresis egrave iacute icircumflex idieresis igrave ntilde oacute ocircumflex odieresis ograve otilde scaron uacute ucircumflex udieresis ugrave yacute ydieresis zcaron exclamsmall Hungarumlautsmall dollaroldstyle dollarsuperior ampersandsmall Acutesmall parenleftsuperior parenrightsuperior twodotenleader onedotenleader zerooldstyle oneoldstyle twooldstyle threeoldstyle fouroldstyle fiveoldstyle sixoldstyle sevenoldstyle eightoldstyle nineoldstyle commasuperior threequartersemdash periodsuperior questionsmall asuperior bsuperior centsuperior dsuperior esuperior isuperior lsuperior msuperior nsuperior osuperior rsuperior ssuperior tsuperior ff ffi ffl parenleftinferior parenrightinferior Circumflexsmall hyphensuperior Gravesmall Asmall Bsmall Csmall Dsmall Esmall Fsmall Gsmall Hsmall Ismall Jsmall Ksmall Lsmall Msmall Nsmall Osmall Psmall Qsmall Rsmall Ssmall Tsmall Usmall Vsmall Wsmall Xsmall Ysmall Zsmall colonmonetary onefitted rupiah Tildesmall exclamdownsmall centoldstyle Lslashsmall Scaronsmall Zcaronsmall Dieresissmall Brevesmall Caronsmall Dotaccentsmall Macronsmall figuredash hypheninferior Ogoneksmall Ringsmall Cedillasmall questiondownsmall oneeighth threeeighths fiveeighths seveneighths onethird twothirds zerosuperior foursuperior fivesuperior sixsuperior sevensuperior eightsuperior ninesuperior zeroinferior oneinferior twoinferior threeinferior fourinferior fiveinferior sixinferior seveninferior eightinferior nineinferior centinferior dollarinferior periodinferior commainferior Agravesmall Aacutesmall Acircumflexsmall Atildesmall Adieresissmall Aringsmall AEsmall Ccedillasmall Egravesmall Eacutesmall Ecircumflexsmall Edieresissmall Igravesmall Iacutesmall Icircumflexsmall Idieresissmall Ethsmall Ntildesmall Ogravesmall Oacutesmall Ocircumflexsmall Otildesmall Odieresissmall OEsmall Oslashsmall Ugravesmall Uacutesmall Ucircumflexsmall Udieresissmall Yacutesmall Thornsmall Ydieresissmall"
	cc = c.split()
	if nn:
		ndict = {}
		nout = []
		nend = []
		for n in nn:
			if n in cc: ndict[n] = 1 # glyph names in CFF order
			else: nend.append(n) # glyph names not in CFF order
		for c in cc:
			if c in ndict: nout.append(c)
		nout.extend(nend)
		return nout
	else: return cc


#---------------------------------------
# Take a glyph list as text (textin),
# return a report of name changes, etc.,
# a feature file, and GOADB
# With namechange False, will retain submitted
# names instead of changing them to prescribed
# work names.
def looklist(textin, d, namechange=True):
	m = [] # output messages
	glyphlist = [] # list to store checked glyph names
	glyphnames = {}
	nn = textin.split()

	for n in nn:
		g = d.glyph(n)
		if g:
			glyphnames[g.name] = n
			# Look for names to change:
			if namechange is True and g.name != n and not g.name in nn:
				glyphlist.append(g.name)
				m.append("Glyph name '%s' changed to '%s'." % (n, g.name))
			else: glyphlist.append(n)
		else: m.append("Glyph %s not found. Skipping." % n)
	glyphlist = cfforder(glyphlist) # sort glyph list by CFF

	# Print GOADB file:
	aliasfile = d.aliasfile(glyphlist)
	print(aliasfile + "\n\n")

	# Print feature file
	featurefile = makefeatures(d, glyphlist)
	print(featurefile + "\n\n")

	# Print derivedchars:
	print(derivedchars(d, glyphlist) + "\n\n")

	# Check for minuscule/majuscule matches:
	for n in glyphlist:
		g = d.glyph(n)
		if g.min and not g.min in glyphnames:
			m.append("Minuscule match %s for glyph %s is missing." % (g.min, n))
		if g.maj and not g.maj in glyphlist:
			m.append("Majuscule match %s for glyph %s is missing." % (g.maj, n))

	return "\n".join(m)


#---------------------------------------
# Confirm a list of glyph names (nn) conforming to Adobe glyph name conventions
# return a list of names which DO NOT conform, unless 'match' is set to True
re_uall = re.compile(r'u(ni)?[0-9A-F]{4}') # Uni-name search pattern
def namecheck(nn, match=False):
	a = agl() # get AGL information
	nout = [] # list of glyph names to return

	if isinstance(nn, str): nn = [nn] # convert a single string to a list
	for n in nn:
		nomatch = 0 # flag for nonconforming names
		nsplit = n.split('.', 1) # separate name at first period
		nparts = nsplit[0].split('_') # separate name root components
		for p in nparts:
			if not (p in a or re_uall.match(p)): nomatch = 1
		if match is False and nomatch: nout.append(n)
		elif match is True and not nomatch: nout.append(n)

	return nout


#-----------------------------------------------------------
# Deconstruct the character mapping of a glyph name (n), and return
# (as a tuple) a list of Unicode values, and a suffix string
re_uni = re.compile(r'^uni((?:[0-9A-F]{4})+)$') # match uni-name
re_u = re.compile(r'^u([0-9A-F]{4}(?:[0-9A-F]{2})?)$') # match u-name
re_code = re.compile(r'[0-9A-F]{4}') # match Unicode value

def namemap(n):
	if not re.compile(r'^[A-Za-z0-9\._]+$').match(n): return None # If the glyph name does not match accepted characters
	uu = [] # collects Unicode values
	namesuffix = None # glyph name suffix

	nameparts = n.split(".", 1)
	if len(nameparts) > 1: namesuffix = nameparts[1] # set name suffix, if any
	nameparts = nameparts[0].split("_")
	for p in nameparts:
		#--- skipped first step to match Zapf Dingbats font names
		if p in aglist: uu.append(aglist[p]) # if name is in the Adobe Glyph List
		elif re_uni.match(p):
			m = re_uni.match(p).group(1)
			uu.extend(re_code.findall(m))
		elif re_u.match(p):
			uu.append(re_u.match(p).group(1))
		else:
			uu.append(None)
	return (uu, namesuffix)

kPunctuation = {
					0x0021 : "exclam",
					0x0022 : "quotation",
					0x0023 : "numbersign",
					0x0026 : "ampersand",
					0x0027 : "apostrophe",
					0x002C : "comma",
					0x002E : "period",
					0x003A : "colon",
					0x003B : "semicolon",
					0x003F : "questionmark",
					0x0040 : "at",
					0x0021 : "questiondown",
					"exclam" : 0x0021,
					"quotation" : 0x0022,
					"numbersign" : 0x0023,
					"ampersand" : 0x0026,
					"apostrophe" : 0x0027,
					"comma" : 0x002C,
					"period" : 0x002E,
					"colon" : 0x003A,
					"semicolon" : 0x003B,
					"questionmark" : 0x003F,
					"at" : 0x0040,
					"questiondown" : 0x0021,
					}

kUnicodeScriptRanges = [
			# Note: CJK ranges omitted. Don;t expect to kern these!
			(0x0000, 0x007F, "Latin", "C0 Controls and Basic Latin (Basic Latin)"),
			(0x0080, 0x00FF, "Latin", "C1 Controls and Latin-1 Supplement (Latin-1 Supplement)"),
			(0x0100, 0x017F, "Latin", "Latin Extended-A"),
			(0x0180, 0x024F, "Latin", "Latin Extended-B"),
			(0x0250, 0x02AF, "IPA", "IPA Extensions"),
			(0x02B0, 0x02FF, "SML", "Spacing Modifier Letters"),
			(0x0300, 0x036F, "CDM", "Combining Diacritical Marks"),
			(0x0370, 0x03FF, "Greek", "Greek and Coptic"),
			(0x0400, 0x04FF, "Cyrillic", "Cyrillic"),
			(0x0500, 0x052F, "Cyrillic", "Cyrillic Supplement"),
			(0x0530, 0x058F, "Armenian", "Armenian"),
			(0x0590, 0x05FF, "Hebrew", "Hebrew"),
			(0x0600, 0x06FF, "Arabic", "Arabic"),
			(0x0700, 0x074F, "Syriac", "Syriac"),
			(0x0750, 0x077F, "Arabic", "Arabic Supplement"),
			(0x0780, 0x07BF, "Thaana", "Thaana"),
			(0x07C0, 0x07FF, "NKo", "NKo"),
			(0x0900, 0x097F, "Devanagari", "Devanagari"),
			(0x0980, 0x09FF, "Bengali", "Bengali"),
			(0x0A00, 0x0A7F, "Gurmukhi", "Gurmukhi"),
			(0x0A80, 0x0AFF, "Gujarati", "Gujarati"),
			(0x0B00, 0x0B7F, "Oriya", "Oriya"),
			(0x0B80, 0x0BFF, "Tamil", "Tamil"),
			(0x0C00, 0x0C7F, "Telugu", "Telugu"),
			(0x0C80, 0x0CFF, "Kannada", "Kannada"),
			(0x0D00, 0x0D7F, "Malayalam", "Malayalam"),
			(0x0D80, 0x0DFF, "Sinhala", "Sinhala"),
			(0x0E00, 0x0E7F, "Thai", "Thai"),
			(0x0E80, 0x0EFF, "Lao", "Lao"),
			(0x0F00, 0x0FFF, "Tibetan", "Tibetan"),
			(0x1000, 0x109F, "Myanmar", "Myanmar"),
			(0x10A0, 0x10FF, "Georgian", "Georgian"),
			(0x1100, 0x11FF, "CJK", "Hangul Jamo"),
			(0x1200, 0x137F, "Ethiopic", "Ethiopic"),
			(0x1380, 0x139F, "Ethiopic", "Ethiopic Supplement"),
			(0x13A0, 0x13FF, "Cherokee", "Cherokee"),
			(0x1400, 0x167F, "UCAS", "Unified Canadian Aboriginal Syllabics"),
			(0x1680, 0x169F, "Ogham", "Ogham"),
			(0x16A0, 0x16FF, "Runic", "Runic"),
			(0x1700, 0x171F, "Tagalog", "Tagalog"),
			(0x1720, 0x173F, "Hanunoo", "Hanunoo"),
			(0x1740, 0x175F, "Buhid", "Buhid"),
			(0x1760, 0x177F, "Tagbanwa", "Tagbanwa"),
			(0x1780, 0x17FF, "Khmer", "Khmer"),
			(0x1800, 0x18AF, "Mongolian", "Mongolian"),
			(0x1900, 0x194F, "Limbu", "Limbu"),
			(0x1950, 0x197F, "TaiLe", "Tai Le"),
			(0x1980, 0x19DF, "TaiLue", "New Tai Lue"),
			(0x19E0, 0x19FF, "Khmer", "Khmer Symbols"),
			(0x1A00, 0x1A1F, "Buginese", "Buginese"),
			(0x1B00, 0x1B7F, "Balinese", "Balinese"),
			(0x1D00, 0x1D7F, "PhoneticExtensions", "Phonetic Extensions"),
			(0x1D80, 0x1DBF, "PhoneticExtensions Supplement", "Phonetic Extensions Supplement"),
			(0x1DC0, 0x1DFF, "CDM", "Combining Diacritical Marks Supplement"),
			(0x1E00, 0x1EFF, "Latin", "Latin Extended Additional"),
			(0x1F00, 0x1FFF, "Greek", "Greek Extended"),
			(0x2000, 0x206F, "Punctuation", "General Punctuation"),
			(0x2070, 0x209F, "SuperscriptsSubscripts", "Superscripts and Subscripts"),
			(0x20A0, 0x20CF, "Currency", "Currency Symbols"),
			(0x20D0, 0x20FF, "CDMS", "Combining Diacritical Marks for Symbols"),
			(0x2100, 0x214F, "LetterlikeSymbols", "Letterlike Symbols"),
			(0x2150, 0x218F, "NumberForms", "Number Forms"),
			(0x2190, 0x21FF, "Arrows", "Arrows"),
			(0x2200, 0x22FF, "Mathematical", "Mathematical Operators"),
			(0x2300, 0x23FF, "MTechnical", "Miscellaneous Technical"),
			(0x2400, 0x243F, "ControlPictures", "Control Pictures"),
			(0x2440, 0x245F, "OCR", "Optical Character Recognition"),
			(0x2460, 0x24FF, "EnclosedAlphanumerics", "Enclosed Alphanumerics"),
			(0x2500, 0x257F, "BoxDrawing", "Box Drawing"),
			(0x2580, 0x259F, "BlockElements", "Block Elements"),
			(0x25A0, 0x25FF, "GeometricShapes", "Geometric Shapes"),
			(0x2600, 0x26FF, "Symbols", "Miscellaneous Symbols"),
			(0x2700, 0x27BF, "Dingbats", "Dingbats"),
			(0x27C0, 0x27EF, "MMathematical", "Miscellaneous Mathematical Symbols-A"),
			(0x27F0, 0x27FF, "Arrows-A", "Supplemental Arrows-A"),
			(0x2800, 0x28FF, "Braille", "Braille Patterns"),
			(0x2900, 0x297F, "Arrows-B", "Supplemental Arrows-B"),
			(0x2980, 0x29FF, "Mathematical", "Miscellaneous Mathematical Symbols-B"),
			(0x2A00, 0x2AFF, "Mathematical", "Supplemental Mathematical Operators"),
			(0x2B00, 0x2BFF, "Symbols", "Miscellaneous Symbols and Arrows"),
			(0x2C00, 0x2C5F, "Glagolitic", "Glagolitic"),
			(0x2C60, 0x2C7F, "Latin", "Latin Extended-C"),
			(0x2C80, 0x2CFF, "Coptic", "Coptic"),
			(0x2D00, 0x2D2F, "Georgian", "Georgian Supplement"),
			(0x2D30, 0x2D7F, "Tifinagh", "Tifinagh"),
			(0x2D80, 0x2DDF, "Ethiopic", "Ethiopic Extended"),
			(0x2E00, 0x2E7F, "Punctuation", "Supplemental Punctuation"),
			(0xA720, 0xA7FF, "Latin", "Latin Extended-D"),
			(0xA800, 0xA82F, "Sylot Nagri", "Syloti Nagri"),
			(0xA840, 0xA87F, "Phags-pa", "Phags-pa"),
			(0xFB00, 0xFB4F, "Alphabetic", "Alphabetic Presentation Forms"),
			(0xFB50, 0xFDFF, "Arabic", "Arabic Presentation Forms-A"),
			(0xFE70, 0xFEFF, "Arabic", "Arabic Presentation Forms-B"),
			(0xFF00, 0xFFEF, "Halfwidth and Fullwidth Forms", "Halfwidth and Fullwidth Forms"),
			(0xFFF0, 0xFFFF, "Specials", "Specials"),
			(0x10000, 0x1007F, "LinearB", "Linear B Syllabary"),
			(0x10080, 0x100FF, "LinearB", "Linear B Ideograms"),
			(0x10100, 0x1013F, "Aegean Numbers", "Aegean Numbers"),
			(0x10140, 0x1018F, "Ancient Greek Numbers", "Ancient Greek Numbers"),
			(0x10300, 0x1032F, "Old Italic", "Old Italic"),
			(0x10330, 0x1034F, "Gothic", "Gothic"),
			(0x10380, 0x1039F, "Ugaritic", "Ugaritic"),
			(0x103A0, 0x103DF, "Old Persian", "Old Persian"),
			(0x10400, 0x1044F, "Deseret", "Deseret"),
			(0x10450, 0x1047F, "Shavian", "Shavian"),
			(0x10480, 0x104AF, "Osmanya", "Osmanya"),
			(0x10800, 0x1083F, "Cypriot", "Cypriot Syllabary"),
			(0x10900, 0x1091F, "Phoenician", "Phoenician"),
			(0x10A00, 0x10A5F, "Kharoshthi", "Kharoshthi"),
			(0x12000, 0x123FF, "Cuneiform", "Cuneiform"),
			(0x12400, 0x1247F, "Cuneiform", "Cuneiform Numbers and Punctuation"),
			(0x1D000, 0x1D0FF, "ByzantineMusicalSymbols", "Byzantine Musical Symbols"),
			(0x1D100, 0x1D1FF, "Musical", "Musical Symbols"),
			(0x1D200, 0x1D24F, "Ancient Greek Musical Notation", "Ancient Greek Musical Notation"),
			(0x1D300, 0x1D35F, "TaiXuanJingSymbols", "Tai Xuan Jing Symbols"),
			(0x1D360, 0x1D37F, "CountingRodNumerals", "Counting Rod Numerals"),
			(0x1D400, 0x1D7FF, "Mathematical", "Mathematical Alphanumeric Symbols"),
						]
#-----------------------------------------------------------
# Get the script for a glyph based on Unicode blocks.
kUnknownScript = "Unknown"
def getscript(UV):
	for entry in kUnicodeScriptRanges:
		if (UV >=entry[0]) and (UV <= entry[1]):
			return entry[2]
	return kUnknownScript



from __future__ import print_function

__copyright__ = """Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
"""

__doc__ = """
stemhist. Wrapper for the Adobe auto-hinting C library. The AC C library
works on only one glyph at a time, expressed in the old 'bez' format, a
bezier-curve language much like PostScript Type 1, but with text
operators and coordinates.

The stemhist script uses Just von Rossum's fontTools library to read and write a
font file. It extracts the PostScript T2 charstring program for each selected
glyph in turn. It converts this into a 'bez language file, and calls the
autohintexe program to work on this file. stemhist returns a returns a list of
either stem hints or alignment zones.


In order to work on Type 1 font files as well as the OpenType/CFF fonts
supported directly by the fontTools library, stemhist uses the 'tx' tool
to convert the font to a 'CFF' font program, and then builds a partial
OpenType font for the fontTools to work with.

"""
__usage__ = """
stemhist program v1.28 Aug 28 2018
stemhist -h
stemhist -u
stemhist [-g <glyph list>] [-gf <filename>] [-xg <glyph list>] [-xgf <filename>] [-all] [-a] [-new] -q font-path1 font-path2...

Stem and Alignment zone report for OT/CFF fonts.
Copyright (c) 2006 Adobe Systems Incorporated
"""

__help__ = __usage__ + """
Examines the specified glyphs, and generates a report of either vertical
and horizontal stems, or alignment zones,  with a count of how many
glyphs have a stem with that zone, and the list of glyphs which have the
zone.

These reports are then used to select alignment zones and horizontal and
vertical stems for use in the global font hinting info. The idea is that
you should either pick the most popular alignment zones for the font
global alignment zone settings, or fix your font so that the important
alignment zones (Cap height, X height, Overshoot, etc), show up as the
most popular alignment zones. Same for stem hints; you should either use
the several most frequent stem widths for the font global stem hint
lists, or fix your font so that the most frequent stem widths are the
important ones.

Note that you can specify a set of glyphs from which to derive your
report. This allows you to get the stem report from just a few glyphs
where the main stems should all be the same, and to compare this with
the report from all the glyphs. If you have done a good job of digital
design, the main stems in all glyphs will lead to a report of a fairly
small set of frequently occurring different stem widths.

Options:

-h	Print help

-u	Print help

-g <glyphID1>,<glyphID2>,...,<glyphIDn>   Report on only the specified
	list of glyphs. The list must be comma-delimited. The glyph ID's may
	be glyphID's, glyph names, or glyph CID's. If the latter, the CID
	value must be prefixed with the string "cid". There must be no
	white-space in the glyph list. Examples: stemhist -g A,B,C,69 myFont
	stemhist -g cid1030,cid34,cid3455,69 myCIDFont

A range of glyphs may be specified by providing two names separated only
	by a hyphen: AC -g zero-nine,onehalf myFont Note that the range will
	be resolved by expanding the glyph indices (GID)'s, not by
	alphabetic names.

-gf <file name>   Report on only the list of glyphs contained in the
	specified file, The file must contain a comma-delimited list of
	glyph identifiers. Any number of space, tab, and new-line characters
	are permitted between glyph names and commas.

-xg, -xgf  Same as -g and -gf, but will exclude the specified glyphs from hinting.

-a   Print alignment zone report rather than stem report

-all Include stems formed by curved line segments; by default, includes
	only stems formed by straight line segments.

-new Skip processing a font if the report files already exist. This
	allows you to use the command: stemhist -new */*/*ps, and apply the
	tool only to fonts where the report has not yet been run.

-o report path
	When this is specified, the path argument is used as the base
	path for the reports. Otherwise, the font file path is used as the
	base path. Several reports are written, using the base path
	name plus an extension.
	Without the -a option, the report files are:
	   <base path >.hstm.txt   The horizontal stems.
	   <base path >.vstm.txt  The vertical stems.
	With the -a option, the report files are:
	   <base path>.top.txt   The top zones.
	  <base path >.bot.txt   The bottom zones.

-q  Don't print glyph names - just print a dot occasionally to show progress.

"""



import sys
import os
import re
import time
from fontTools.ttLib import TTFont, getTableModule
from fontTools.misc.py23 import open, tounicode
from afdko.autohint import (
	parseGlyphListArg, getGlyphID, getGlyphNames, filterGlyphList, openFile,
	ACOptionParseError, ACFontError, logMsg, ACreport, expandNames,
	AUTOHINTEXE)
from afdko import fdkutils, ufotools
import traceback
from collections import defaultdict


rawData = []


class ACOptions:
	def __init__(self):
		self.fileList = []
		self.glyphList = []
		self.excludeGlyphList = 0
		self.allStems = 0
		self.doAlign = 0
		self.reportPath = None
		self.new = 0
		self.noFlex = 1
		self.allow_no_blues = 1
		self.vCounterGlyphs = []
		self.hCounterGlyphs = []
		self.verbose = 1
		self.debug = 0


def getOptions():
	options = ACOptions()
	i = 1
	numOptions = len(sys.argv)
	while i < numOptions:
		arg = sys.argv[i]
		if options.fileList and arg[0] == "-":
			raise ACOptionParseError("Option Error: All file names must follow all other options <%s>." % arg)

		if arg == "-h":
			logMsg(__help__)
			sys.exit(0)
		elif arg == "-u":
			logMsg(__usage__)
			sys.exit(0)
		elif arg == "-all":
			options.allStems = 1
		elif arg == "-a":
			options.doAlign = 1
		elif arg == "-d":
			options.debug = 1
		elif arg == "-q":
			options.verbose = 0
		elif arg == "-o":
			i = i +1
			options.reportPath = sys.argv[i]
		elif arg == "-new":
			options.new = 1
		elif arg in ["-xg", "-g"]:
			if arg == "-xg":
				options.excludeGlyphList = 1
			i = i +1
			glyphString = sys.argv[i]
			if glyphString[0] == "-":
				raise ACOptionParseError("Option Error: it looks like the first item in the glyph list following '-g' is another option.")
			options.glyphList += parseGlyphListArg(glyphString)
		elif arg in ["-xgf", "-gf"]:
			if arg == "-xgf":
				options.excludeGlyphList = 1
			i = i +1
			filePath = sys.argv[i]
			if filePath[0] == "-":
				raise ACOptionParseError("Option Error: it looks like the the glyph list file following '-gf' is another option.")
			try:
				with open(filePath, "r", encoding='utf-8') as gf:
					glyphString = gf.read()
			except (IOError,OSError):
				raise ACOptionParseError("Option Error: could not open glyph list file <%s>." %  filePath)
			options.glyphList += parseGlyphListArg(glyphString)
		elif arg[0] == "-":
			raise ACOptionParseError("Option Error: Unknown option <%s>." %  arg)
		else:
			arg = arg.rstrip(os.sep) # might be a UFO font. auto completion in some shells adds a dir separator, which then causes problems with os.path.dirname().
			options.fileList += [arg]
		i  += 1

	if not options.fileList:
		raise ACOptionParseError("Option Error: You must provide at least one font file path.")

	return options


kCharZone = "charZone"
kStemZone = "stemZone"
kHStem = "HStem"
kVStem = "VStem"
class GlyphReports:
	def __init__(self):
		self.glyphName = None
		self.hStemList = {}
		self.vStemList = {}
		self.hStemPosList = {}
		self.vStemPosList = {}
		self.charZoneList = {}
		self.stemZoneStemList = {}
		self.glyphs = {}

	def startGlyphName(self, glyphName):
		self.hStemList = {}
		self.vStemList = {}
		self.hStemPosList = {}
		self.vStemPosList = {}
		self.charZoneList = {}
		self.stemZoneStemList = {}
		self.glyphs[glyphName] = [self.hStemList, self.vStemList, self.charZoneList, self.stemZoneStemList]
		self.glyphName = glyphName

	def addGlyphReport(self, reportString):
		lines = reportString.splitlines()
		for line in lines:
			tokenList = line.split()
			key = tokenList[0]
			x = eval(tokenList[3])
			y = eval(tokenList[5])
			hintpos = "%s %s" % (x ,y)
			if key == kCharZone:
				self.charZoneList[hintpos] = (x, y)
			elif key == kStemZone:
				self.stemZoneStemList[hintpos] = (x, y)
			elif key == kHStem:
				width = x - y
				if hintpos not in self.hStemPosList: # avoid counting duplicates
					count = self.hStemList.get(width, 0)
					self.hStemList[width] = count+1
					self.hStemPosList[hintpos] = width
			elif key == kVStem:
				width = x - y
				if hintpos not in self.vStemPosList: # avoid counting duplicates
					count = self.vStemList.get(width, 0)
					self.vStemList[width] = count+1
					self.vStemPosList[hintpos] = width
			else:
				raise ACFontError("Error: Found unknown keyword %s in report file for glyph %s." % (key, self.glyphName))

	@staticmethod
	def round_value(val):
		if val >= 0:
			return int(val + 0.5)
		else:
			return int(val - 0.5)

	def parse_stem_dict(self, stem_dict):
		"""
		stem_dict: {45.5: 1, 47.0: 2}
		"""
		# key: stem width
		# value: stem count
		width_dict = defaultdict(int)
		for width, count in stem_dict.items():
			width = self.round_value(width)
			width_dict[width] += count
		return width_dict

	def parse_zone_dicts(self, char_dict, stem_dict):
		all_zones_dict = char_dict.copy()
		all_zones_dict.update(stem_dict)
		# key: zone height
		# value: zone count
		top_dict = defaultdict(int)
		bot_dict = defaultdict(int)
		for top, bot in all_zones_dict.values():
			top = self.round_value(top)
			top_dict[top] += 1
			bot = self.round_value(bot)
			bot_dict[bot] += 1
		return top_dict, bot_dict

	@staticmethod
	def assemble_rep_list(items_dict, count_dict):
		# item 0: stem/zone count
		# item 1: stem width/zone height
		# item 2: list of glyph names
		rep_list = []
		for item in items_dict:
			rep_list.append((count_dict[item], item, sorted(items_dict[item])))
		return rep_list

	def getReportLists(self):
		"""
		self.glyphs is a dictionary:
			key: glyph name
			value: list of 4 dictionaries
					self.hStemList
					self.vStemList
					self.charZoneList
					self.stemZoneStemList
		{
		 'A': [{45.5: 1, 47.0: 2}, {229.0: 1}, {}, {}],
		 'B': [{46.0: 2, 46.5: 2, 47.0: 1}, {94.0: 1, 95.0: 1, 100.0: 1}, {}, {}],
		 'C': [{50.0: 2}, {109.0: 1}, {}, {}],
		 'D': [{46.0: 1, 46.5: 2, 47.0: 1}, {95.0: 1, 109.0: 1}, {}, {}],
		 'E': [{46.5: 2, 47.0: 1, 50.0: 2, 177.0: 1, 178.0: 1},
		       {46.0: 1, 75.5: 2, 95.0: 1}, {}, {}],
		 'F': [{46.5: 2, 47.0: 1, 50.0: 1, 177.0: 1},
		       {46.0: 1, 60.0: 1, 75.5: 1, 95.0: 1}, {}, {}],
		 'G': [{43.0: 1, 44.5: 1, 50.0: 1, 51.0: 1}, {94.0: 1, 109.0: 1}, {}, {}]
		}
		"""
		h_stem_items_dict = defaultdict(set)
		h_stem_count_dict = defaultdict(int)
		v_stem_items_dict = defaultdict(set)
		v_stem_count_dict = defaultdict(int)

		top_zone_items_dict = defaultdict(set)
		top_zone_count_dict = defaultdict(int)
		bot_zone_items_dict = defaultdict(set)
		bot_zone_count_dict = defaultdict(int)

		for gName, dicts in self.glyphs.items():
			hStemDict, vStemDict, charZoneDict, stemZoneStemDict = dicts

			glyph_h_stem_dict = self.parse_stem_dict(hStemDict)
			glyph_v_stem_dict = self.parse_stem_dict(vStemDict)

			for stem_width, stem_count in glyph_h_stem_dict.items():
				h_stem_items_dict[stem_width].add(gName)
				h_stem_count_dict[stem_width] += stem_count

			for stem_width, stem_count in glyph_v_stem_dict.items():
				v_stem_items_dict[stem_width].add(gName)
				v_stem_count_dict[stem_width] += stem_count

			glyph_top_zone_dict, glyph_bot_zone_dict = self.parse_zone_dicts(
				charZoneDict, stemZoneStemDict)

			for zone_height, zone_count in glyph_top_zone_dict.items():
				top_zone_items_dict[zone_height].add(gName)
				top_zone_count_dict[zone_height] += zone_count

			for zone_height, zone_count in glyph_bot_zone_dict.items():
				bot_zone_items_dict[zone_height].add(gName)
				bot_zone_count_dict[zone_height] += zone_count

		# item 0: stem count
		# item 1: stem width
		# item 2: list of glyph names
		h_stem_list = self.assemble_rep_list(
			h_stem_items_dict, h_stem_count_dict)

		v_stem_list = self.assemble_rep_list(
			v_stem_items_dict, v_stem_count_dict)

		# item 0: zone count
		# item 1: zone height
		# item 2: list of glyph names
		top_zone_list = self.assemble_rep_list(
			top_zone_items_dict, top_zone_count_dict)

		bot_zone_list = self.assemble_rep_list(
			bot_zone_items_dict, bot_zone_count_dict)

		return h_stem_list, v_stem_list, top_zone_list, bot_zone_list


def srtCnt():
	"""
	sort by: count (1st item), value (2nd item), list of glyph names (3rd item)
	"""
	return lambda t: (-t[0], -t[1], t[2])


def srtVal():
	"""
	sort by: value (2nd item), count (1st item), list of glyph names (3rd item)
	"""
	return lambda t: (t[1], -t[0], t[2])


def srtRevVal():
	"""
	sort by: value (2nd item), count (1st item), list of glyph names (3rd item)
	"""
	return lambda t: (-t[1], -t[0], t[2])


def formatReport(rep_list, sortFunc):
	rep_list.sort(key=sortFunc())
	return ["%s\t%s\t%s\n" % (item[0], item[1], item[2]) for item in rep_list]


def checkReportExists(path, doAlign):
	if doAlign:
		suffixes = (".top.txt", ".bot.txt")
	else:
		suffixes = (".hstm.txt", ".vstm.txt")
	foundOne = 0
	for i in range(len(suffixes)):
		fName = path + suffixes[i]
		if os.path.exists(fName):
			foundOne = 1
			break
	return foundOne


def PrintReports(path, h_stem_list, v_stem_list, top_zone_list, bot_zone_list):
	items = ([h_stem_list, srtCnt], [v_stem_list, srtCnt],
		     [top_zone_list, srtRevVal], [bot_zone_list, srtVal])
	atime = time.asctime()
	suffixes = (".hstm.txt", ".vstm.txt", ".top.txt", ".bot.txt")
	titles = ("Horizontal Stem List for %s on %s\n" % (path, atime),
				"Vertical Stem List for %s on %s\n" % (path, atime),
				"Top Zone List for %s on %s\n" % (path, atime),
				"Bottom Zone List for %s on %s\n" % (path, atime),
			)
	headers = ("Count\tWidth\tGlyph List\n",
			"Count\tWidth\tGlyph List\n",
			"Count\tTop Zone\tGlyph List\n",
			"Count\tBottom Zone\tGlyph List\n",
			)
	for i, item in enumerate(items):
		rep_list, sortFunc = item
		if not rep_list:
			continue
		fName = '{}{}'.format(path, suffixes[i])
		title = titles[i]
		header = headers[i]
		try:
			with open(fName, "w") as fp:
				fp.write(tounicode(title))
				fp.write(tounicode(header))
				for item in formatReport(rep_list, sortFunc):
					fp.write(tounicode(item))
				print("Wrote %s" % fName)
		except (IOError, OSError):
			print("Error creating file %s!" % fName)


def collectStemsFont(path, options):
	#    use fontTools library to open font and extract CFF table.
	#    If error, skip font and report error.
	fontFileName = os.path.basename(path)
	logMsg("")
	if options.doAlign:
		logMsg( "Collecting alignment zones for font %s. Start time: %s." % (path, time.asctime()))
	else:
		logMsg( "Collecting stems for font %s. Start time: %s." % (path, time.asctime()))

	try:
		fontData = openFile(path)
	except (IOError, OSError):
		logMsg( traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1])[-1])
		raise ACFontError("Error opening or reading from font file <%s>." % fontFileName)
	except:
		logMsg( traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1])[-1])
		raise ACFontError("Error parsing font file <%s>." % fontFileName)

	#   filter specified list, if any, with font list.
	fontGlyphList = fontData.getGlyphList()
	glyphList = filterGlyphList(options, fontGlyphList, fontFileName)
	if not glyphList:
		raise ACFontError("Error: selected glyph list is empty for font <%s>." % fontFileName)

	tempBez = fdkutils.get_temp_file_path()
	tempReport = '{}{}'.format(tempBez, ".rpt")
	tempFI = fdkutils.get_temp_file_path()

	#    open font plist file, if any. If not, create empty font plist.
	psName = fontData.getPSName()

	#    build alignment zone string
	allow_no_blues =  1
	noFlex = 0
	vCounterGlyphs = hCounterGlyphs = []
	fdGlyphDict, fontDictList = fontData.getfdInfo(psName, path, allow_no_blues, noFlex, vCounterGlyphs, hCounterGlyphs, glyphList)

	if fdGlyphDict == None:
		fdDict = fontDictList[0]
		with open(tempFI, "w") as fp:
			fp.write(tounicode(fdDict.getFontInfo()))
	else:
		if not options.verbose:
			logMsg("Note: Using alternate FDDict global values from fontinfo file for some glyphs. Remove option '-q' to see which dict is used for which glyphs.")

	removeHints = 1
	isCID = fontData.isCID()
	lastFDIndex = None
	glyphReports = GlyphReports()

	if not options.verbose:
		dotCount = 0
		curTime = time.time()

	for name in glyphList:
		if name == ".notdef":
			continue

		if options.verbose:
			logMsg("Checking %s." %name)
		else:
			newTime = time.time()
			if (newTime - curTime) > 1:
				print(".", end=' ')
				sys.stdout.flush()
				curTime = newTime
				dotCount +=1
			if dotCount > 40:
				dotCount = 0
				print("")

		# 	Convert to bez format
		bezString, width, hasHints = fontData.convertToBez(name, removeHints, options.verbose)
		if bezString == None:
			continue
		if "mt" not in bezString:
			# skip empty glyphs.
			continue

		# get new fontinfo string if FD array index has changed, as
		# as each FontDict has different alignment zones.
		gid = fontData.getGlyphID(name)
		if isCID: #
			fdIndex = fontData.getfdIndex(gid)
			if not fdIndex == lastFDIndex:
				lastFDIndex = fdIndex
				fdDict = fontData.getFontInfo(psName, path, options.allow_no_blues, options.noFlex, options.vCounterGlyphs, options.hCounterGlyphs, fdIndex)
				with open(tempFI, "w") as fp:
					fp.write(tounicode(fdDict.getFontInfo()))
		else:
			if (fdGlyphDict != None):
				try:
					fdIndex = fdGlyphDict[name][0]
				except KeyError:
					# use default dict.
					fdIndex = 0
				if lastFDIndex != fdIndex:
					lastFDIndex = fdIndex
					fdDict = fontDictList[fdIndex]
					with open(tempFI, "w") as fp:
						fp.write(tounicode(fdDict.getFontInfo()))

		glyphReports.startGlyphName(name)


		# 	Call auto-hint library on bez string.
		with open(tempBez, "w") as bp:
			bp.write(tounicode(bezString))

		if options.doAlign:
			doAlign = "-ra"
		else:
			doAlign = "-rs"

		if options.allStems:
			allStems = "-a"
		else:
			allStems = ""

		command = AUTOHINTEXE + " -q %s %s  -f \"%s\" \"%s\" 2>&1" % (doAlign, allStems, tempFI, tempBez)
		if options.debug:
			print(command)
		log = fdkutils.runShellCmd(command)
		if log:
			print(log)
			if "number terminator while" in log:
				print(tempBez)
				sys.exit()

		if os.path.exists(tempReport):
			with open(tempReport, "r", encoding='utf-8') as bp:
				report = bp.read()
			if options.debug:
				print("Wrote AC fontinfo data file to", tempFI)
				print("Wrote AC output rpt file to", tempReport)
			report.strip()
			if report:
				glyphReports.addGlyphReport(report)
				if options.debug:
					rawData.append(report)
		else:
			print("Error - failure in processing outline data")
			report = None

	h_stem_list, v_stem_list, top_zone_list, bot_zone_list = glyphReports.getReportLists()
	if options.reportPath:
		reportPath = options.reportPath
	else:
		reportPath = path
	PrintReports(reportPath, h_stem_list, v_stem_list, top_zone_list, bot_zone_list)
	fontData.close()
	logMsg( "Done with font %s. End time: %s." % (path, time.asctime()))


def main():
	try:
		options = getOptions()
	except ACOptionParseError as e:
		logMsg(e)
		return 1

	# verify that all files exist.
	haveFiles = True
	for path in options.fileList:
		if not os.path.exists(path):
			logMsg( "File does not exist: <%s>." % path)
			haveFiles = False
	if not haveFiles:
		return 1

	for path in options.fileList:
		if options.new:
			if options.reportPath:
				reportPath = options.reportPath
			else:
				reportPath = path
			foundOne = checkReportExists(reportPath, options.doAlign)
			if foundOne:
				logMsg( "Skipping %s, as a report already exists." % (path))
				continue
		try:
			collectStemsFont(path, options)
		except (ACFontError, ufotools.UFOParseError) as e:
			logMsg( "\t%s" % e)
			return 1
		if options.debug:
			with open("rawdata.txt", "w") as fp:
				fp.write(tounicode('\n'.join(rawData)))


if __name__=='__main__':
	sys.exit(main())



from __future__ import print_function
import sys

"""
beztools.py v 1.15 Oct 10 2018

Utilities for converting between T2 charstrings and the bez data format.
Used by autohint and checkoutlines.
"""
__copyright__ = """Copyright 2014-2017 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
"""

import re
import time
import os
from afdko import fdkutils, convertfonttocid
from fontTools.misc.psCharStrings import T2OutlineExtractor, SimpleT2Decompiler
from fontTools.misc.py23 import byteord, bytechr
from fontTools.pens.basePen import BasePen
debug = 0
def debugMsg(*args):
	if debug:
		print(args)
kStackLimit = 46
kStemLimit = 96

class ACFontError(Exception):
	pass

class SEACError(Exception):
	pass

def hintOn( i, hintMaskBytes):
	# used to add the active hints to the bez string, when a  T2 hintmask operator is encountered.
	byteIndex = i // 8
	byteValue = byteord(hintMaskBytes[byteIndex])
	offset = 7 -  (i %8)
	return ((2**offset) & byteValue) > 0


class T2ToBezExtractor(T2OutlineExtractor):
	# The T2OutlineExtractor class calls a class method as the handler for each T2 operator.
	# I use this to convert the T2 operands and arguments to bez operators.
	# Note: flex is converted to regulsr rcurveto's.
	# cntrmasks just map to hint replacement blocks with the specified stems.
	def __init__(self, localSubrs, globalSubrs, nominalWidthX, defaultWidthX, removeHints = 0, allowDecimals = 0):
		T2OutlineExtractor.__init__(self, None, localSubrs, globalSubrs, nominalWidthX, defaultWidthX)
		self.vhints = []
		self.hhints = []
		self.bezProgram = []
		self.removeHints = removeHints
		self.firstMarkingOpSeen = 0
		self.closePathSeen = 0
		self.subrLevel = 0
		self.allowDecimals = allowDecimals

	def execute(self, charString):
		self.subrLevel += 1
		SimpleT2Decompiler.execute(self,charString)
		self.subrLevel -= 1
		if (not self.closePathSeen) and (self.subrLevel == 0):
			 self.closePath()

	def rMoveTo(self, point):
		point = self._nextPoint(point)
		if not self.firstMarkingOpSeen :
			self.firstMarkingOpSeen = 1
			self.bezProgram.append("sc\n")
		debugMsg("moveto", point, "curpos", self.currentPoint)
		x = point[0]
		y = point[1]
		if (not self.allowDecimals):
			x = int(round(x))
			y = int(round(y))
			self.bezProgram.append("%s  %s mt\n" % (x, y))
		else:
			self.bezProgram.append("%.2f  %.2f mt\n" % (x, y))
		self.sawMoveTo = 1

	def rLineTo(self, point):
		point = self._nextPoint(point)
		if not self.firstMarkingOpSeen :
			self.firstMarkingOpSeen = 1
			self.bezProgram.append("sc\n")
			self.bezProgram.append("0 0 mt\n")
		debugMsg("lineto", point, "curpos", self.currentPoint)
		if not self.sawMoveTo:
			self.rMoveTo((0, 0))
		x = point[0]
		y = point[1]
		if (not self.allowDecimals):
			x = int(round(x))
			y = int(round(y))
			self.bezProgram.append("%s  %s dt\n" % (x, y))
		else:
			self.bezProgram.append("%.2f  %.2f dt\n" % (x, y))

	def rCurveTo(self, pt1, pt2, pt3):
		pt1 = list(self._nextPoint(pt1))
		pt2 = list(self._nextPoint(pt2))
		pt3 = list(self._nextPoint(pt3))
		if not self.firstMarkingOpSeen :
			self.firstMarkingOpSeen = 1
			self.bezProgram.append("sc\n")
			self.bezProgram.append("0 0 mt\n")
		debugMsg("curveto", pt1, pt2, pt3, "curpos", self.currentPoint)
		if not self.sawMoveTo:
			self.rMoveTo((0, 0))
		if (not self.allowDecimals):
			for pt in [pt1, pt2, pt3]:
				pt[0] = int(round(pt[0]))
				pt[1] = int(round(pt[1]))
			self.bezProgram.append("%s %s %s %s %s %s ct\n" %  (pt1[0],  pt1[1],  pt2[0],  pt2[1], pt3[0],  pt3[1]))
		else:
			self.bezProgram.append("%.2f %.2f %.2f %.2f %.2f %.2f ct\n" %  (pt1[0],  pt1[1],  pt2[0],  pt2[1], pt3[0],  pt3[1]))

	def op_endchar(self, index):
		self.endPath()
		args = self.popallWidth()
		if args: #  It is a 'seac' composite character. Don't process
			raise SEACError

	def endPath(self):
		# In T2 there are no open paths, so always do a closePath when
		# finishing a sub path.
		if self.sawMoveTo:
			debugMsg("endPath")
			self.bezProgram.append("cp\n")
		self.sawMoveTo = 0
	def closePath(self):
		self.closePathSeen = 1
		debugMsg("closePath")
		if self.bezProgram and self.bezProgram[-1] != "cp\n":
			self.bezProgram.append("cp\n")
		self.bezProgram.append("ed\n")

	def updateHints(self, args,  hintList, bezCommand, writeHints = 1):
		self.countHints(args)

		# first hint value is absolute hint coordinate, second is hint width
		if self.removeHints:
			writeHints = 0

		if not writeHints:
			return

		lastval = args[0]
		if isinstance(lastval, int):
			lastval = float(lastval)/0x10000
			arg = str(lastval) + " 100 div"
		else:
			arg = str(lastval)
		hintList.append(arg)
		self.bezProgram.append(arg)

		for i in list(range(len(args))[1:]):
			val = args[i]
			if isinstance(val, int):
				val = float(val)/0x10000
			newVal =  lastval + val
			lastval = newVal

			if i % 2:
				if isinstance(val, float):
					if val % 1 != 0:
						arg = str(int(val*100)) + " 100 div"
					else:
						arg = str(int(val))
				else:
					arg = str(val)
				hintList.append( arg )
				self.bezProgram.append(arg)
				self.bezProgram.append(bezCommand + "\n")
			else:
				if isinstance(newVal, float):
					if newVal % 1 != 0:
						arg = str(int(newVal*100)) + " 100 div"
					else:
						arg = str(int(newVal))
				else:
					arg = str(newVal)
				hintList.append( arg )
				self.bezProgram.append(arg)


	def op_hstem(self, index):
		args = self.popallWidth()
		self.hhints = []
		self.updateHints(args, self.hhints, "rb")
		debugMsg("hstem", self.hhints)

	def op_vstem(self, index):
		args = self.popallWidth()
		self.vhints = []
		self.updateHints(args, self.vhints, "ry")
		debugMsg("vstem", self.vhints)

	def op_hstemhm(self, index):
		args = self.popallWidth()
		self.hhints = []
		self.updateHints(args, self.hhints, "rb")
		debugMsg("stemhm", self.hhints, args)

	def op_vstemhm(self, index):
		args = self.popallWidth()
		self.vhints = []
		self.updateHints(args, self.vhints, "ry")
		debugMsg("vstemhm", self.vhints, args)

	def getCurHints(self, hintMaskBytes):
		curhhints = []
		curvhints = []
		numhhints = len(self.hhints)

		for i in range(numhhints//2):
			if hintOn(i, hintMaskBytes):
				curhhints.extend(self.hhints[2*i:2*i+2])
		numvhints = len(self.vhints)
		for i in range(numvhints//2):
			if hintOn(i + numhhints//2, hintMaskBytes):
				curvhints.extend(self.vhints[2*i:2*i+2])
		return curhhints, curvhints

	def doMask(self, index, bezCommand):
		args = []
		if not self.hintMaskBytes:
			args = self.popallWidth()
			if args:
				self.vhints = []
				self.updateHints(args, self.vhints, "ry")
			self.hintMaskBytes = (self.hintCount + 7) // 8

		self.hintMaskString, index = self.callingStack[-1].getBytes(index, self.hintMaskBytes)

		if not  self.removeHints:
			curhhints, curvhints = self.getCurHints( self.hintMaskString)
			strout = ""
			mask = [strout + hex(byteord(ch)) for ch in self.hintMaskString]
			debugMsg(bezCommand, mask, curhhints, curvhints, args)

			self.bezProgram.append("beginsubr snc\n")
			i = 0
			for hint in curhhints:
				self.bezProgram.append(str(hint))
				if i %2:
					self.bezProgram.append("rb\n")
				i +=1
			i = 0
			for hint in curvhints:
				self.bezProgram.append(str(hint))
				if i %2:
					self.bezProgram.append("ry\n")
				i +=1
			self.bezProgram.extend(["endsubr enc\n", "newcolors\n"])
		return self.hintMaskString, index

	def op_hintmask(self, index):
		hintMaskString, index =  self.doMask(index, "hintmask")
		return hintMaskString, index

	def op_cntrmask(self, index):
		hintMaskString, index =   self.doMask(index, "cntrmask")
		return hintMaskString, index


	def countHints(self, args):
		self.hintCount = self.hintCount + len(args) // 2

def convertT2GlyphToBez(t2CharString, removeHints = 0, allowDecimals = 0):
	# wrapper for T2ToBezExtractor which applies it to the supplied T2 charstring
	bezString = ""
	subrs = getattr(t2CharString.private, "Subrs", [])
	extractor = T2ToBezExtractor(subrs, t2CharString.globalSubrs,
				t2CharString.private.nominalWidthX, t2CharString.private.defaultWidthX, removeHints, allowDecimals)
	extractor.execute(t2CharString)
	if extractor.gotWidth:
		t2Wdth = extractor.width - t2CharString.private.nominalWidthX
	else:
		t2Wdth = None
	return "".join(extractor.bezProgram), extractor.hintCount > 0, t2Wdth

class HintMask:
	# class used to collect hints for the current hint mask when converting bez to T2.
	def __init__(self, listPos):
		self.listPos = listPos # The index into the t2list is kept so we can quickly find them later.
		self.hList = [] # These contain the actual hint values.
		self.vList = []

	def maskByte(self, hHints, vHints):
		# return hintmask bytes for known hints.
		numHHints = len(hHints)
		numVHints = len(vHints)
		maskVal = 0
		byteIndex = 0
		self.byteLength = byteLength = int((7 + numHHints + numVHints)/8)
		mask = b""
		self.hList.sort()
		for hint in self.hList:
			try:
				i = hHints.index(hint)
			except ValueError:
				continue	# we get here if some hints have been dropped because of the stack limit
			newbyteIndex = (i//8)
			if newbyteIndex != byteIndex:
				mask += bytechr(maskVal)
				byteIndex +=1
				while byteIndex < newbyteIndex:
					mask += b"\0"
					byteIndex +=1
				maskVal = 0
			maskVal += 2**(7 - (i %8))

		self.vList.sort()
		for hint in self.vList:
			try:
				i = numHHints + vHints.index(hint)
			except ValueError:
				continue	# we get here if some hints have been dropped because of the stack limit
			newbyteIndex = (i//8)
			if newbyteIndex != byteIndex:
				mask += bytechr(maskVal)
				byteIndex +=1
				while byteIndex < newbyteIndex:
					mask += b"\0"
					byteIndex +=1
				maskVal = 0
			maskVal += 2**(7 - (i %8))

		if maskVal:
			mask += bytechr(maskVal)

		if len(mask) < byteLength:
			mask += b"\0"*(byteLength - len(mask))
		self.mask = mask
		return mask

def makeHintList(hints, needHintMasks, isH):
	# Add the list of T2 tokens that make up the initial hint operators
	hintList = []
	lastPos = 0
	# In bez terms, the first coordinate in each pair is absolute, second is relative.
	# In T2, each term is relative to the previous one.
	for hint in hints:
		if not hint:
			continue
		pos1 = hint[0]
		pos = pos1 - lastPos
		if pos % 1 == 0:
			pos = int(pos)
		hintList.append(pos)
		pos2 = hint[1]
		if pos2 % 1 == 0:
			pos2 = int(pos2)
		lastPos = pos1 + pos2
		hintList.append(pos2)

	if  needHintMasks:
		if isH:
			op = "hstemhm"
			hintList.append(op)
		# never need to append vstemhm: if we are using it, it is followed
		# by a mask command and vstemhm is inferred.
	else:
		if isH:
			op = "hstem"
		else:
			op = "vstem"
		hintList.append(op)

	return hintList


bezToT2 = {
		"mt" : 'rmoveto',
		"rmt" : 'rmoveto',
		 "hmt" : 'hmoveto',
		"vmt" : 'vmoveto',
		"dt" : 'rlineto',
		"rdt" : 'rlineto',
		"hdt" : "hlineto",
		"vdt" : "vlineto",
		"ct" : 'rrcurveto',
		"rct" : 'rrcurveto',
		"rcv" : 'rrcurveto',  # Morisawa's alternate name for 'rct'.
		"vhct": 'vhcurveto',
		"hvct": 'hvcurveto',
		"cp" : "",
		"ed" : 'endchar'
		}


def optimizeT2Program(t2List):
	# Assumes T2 operands are in a list with one entry per operand, and each entry is a list of [argList, opToken].
	# Matches logic in tx, and Adobe low level library.
	# Note that I am expecting only rlineto,vlinteto,hlinto, vhcurveto,hvcurveto,rcurveto froom AC.
	# The other opimtized operators, specifically the line-curve combinations are not supported here.
	newT2List = []
	arglist = []
	kNoOp = "noop"
	pendingOp = kNoOp
	sequenceOp = kNoOp
	for entry in t2List:
		op =  entry[1]
		args = entry[0]

		if op == "vlineto":
			dy = args[-1]
			if  (pendingOp in ["vlineto", "hlineto"]) and (sequenceOp ==  "hlineto"):
				arglist.append(dy)
				sequenceOp = "vlineto"
				if len(arglist) >= kStackLimit:
					newT2List.append([arglist[:-1], pendingOp])
					arglist = [dy]
					pendingOp = "vlineto"

			else:
				if  pendingOp != kNoOp:
					newT2List.append([arglist, pendingOp])
				arglist = [dy]
				pendingOp = sequenceOp = "vlineto"

		elif op ==  "hlineto":
			dx = args[-1]
			if  (pendingOp in ["vlineto", "hlineto"]) and (sequenceOp ==  "vlineto"):
				arglist.append(dx)
				sequenceOp = "hlineto"
				if len(arglist) >= kStackLimit:
					newT2List.append([arglist[:-1], pendingOp])
					arglist = [dx]
					pendingOp = "hlineto"
			else:
				if  pendingOp != kNoOp:
					newT2List.append([arglist, pendingOp])
				arglist = [dx]
				pendingOp = sequenceOp = "hlineto"

		elif op == "rlineto":
			dx = args[-2]
			dy = args[-1]
			if dx == 0:
				if  (pendingOp in ["vlineto", "hlineto"]) and (sequenceOp ==  "hlineto"):
					arglist.append(dy)
					sequenceOp =  "vlineto"
					if len(arglist) >= kStackLimit:
						newT2List.append([arglist[:-1], pendingOp])
						arglist = [dy]
						pendingOp = "vlineto"
				else:
					if  pendingOp != kNoOp:
						newT2List.append([arglist, pendingOp])
					arglist = [dy]
					pendingOp = sequenceOp = "vlineto"

			elif dy == 0:
				if  (pendingOp in ["vlineto", "hlineto"]) and (sequenceOp ==  "vlineto"):
					arglist.append(dx)
					sequenceOp =  "hlineto"
					if len(arglist) >= kStackLimit:
						newT2List.append([arglist[:-1], pendingOp])
						arglist = [dx]
						pendingOp = "hlineto"
				else:
					if  pendingOp != kNoOp:
						newT2List.append([arglist, pendingOp])
					arglist = [dx]
					pendingOp = sequenceOp = "hlineto"

			elif pendingOp == "rrcurveto":
				arglist.extend([dx,dy])

				if len(arglist) >= kStackLimit:
					newT2List.append([arglist[:-2], pendingOp])
					arglist = [dx, dy]
					pendingOp = sequenceOp = "rlineto"
				else:
					newT2List.append([arglist, "rcurveline"])
					arglist = []
					pendingOp = sequenceOp = kNoOp


			elif  (pendingOp == op) and (sequenceOp == op):
				arglist.extend([dx,dy])
				if len(arglist) >= kStackLimit:
					newT2List.append([arglist[:-2], pendingOp])
					arglist = [dx, dy]

			else:
				if pendingOp != kNoOp:
					newT2List.append([arglist, pendingOp])
				arglist = [dx,dy]
				pendingOp = sequenceOp = op


		elif op == "vhcurveto":
			if  (pendingOp in ["vhcurveto", "hvcurveto"]) and (sequenceOp ==  "hvcurveto"):
				sequenceOp = "vhcurveto"
				arglist.extend(args)
				if len(arglist) >= kStackLimit:
					newT2List.append([arglist[:-len(args)], pendingOp])
					arglist = args
					pendingOp = sequenceOp = op
			else:
				if  pendingOp != kNoOp:
					newT2List.append([arglist, pendingOp])
				arglist = args
				pendingOp = sequenceOp = "vhcurveto"
			if len(args) == 5:
				newT2List.append([arglist, pendingOp])
				arglist = []
				pendingOp = sequenceOp = kNoOp


		elif op == "hvcurveto":
			if  (pendingOp in ["vhcurveto", "hvcurveto"]) and (sequenceOp ==  "vhcurveto"):
				sequenceOp = "hvcurveto"
				arglist.extend(args)
				if len(arglist) >= kStackLimit:
					newT2List.append([arglist[:-len(args)], pendingOp])
					arglist = args
					pendingOp = sequenceOp = op
			else:
				if  pendingOp != kNoOp:
					newT2List.append([arglist, pendingOp])
				arglist = args
				pendingOp = sequenceOp = "hvcurveto"
			if len(args) == 5:
				newT2List.append([arglist, pendingOp])
				arglist = []
				pendingOp = sequenceOp = kNoOp


		elif op == "rrcurveto":
			dx1 = args[0]
			dy1 = args[1]
			dx2 = args[2]
			dy2 = args[3]
			dx3 = args[4]
			dy3 = args[5]

			if dx1 == 0:
				if dy3 == 0: # - dy1 dx2 dy2 dx3 - vhcurveto
					if  (pendingOp in ["vhcurveto", "hvcurveto"]) and (sequenceOp ==  "hvcurveto"):
						arglist.extend( [dy1, dx2, dy2, dx3])
						sequenceOp = "vhcurveto"
						if len(arglist) >= kStackLimit:
							newT2List.append([arglist[:-4], pendingOp])
							arglist = [dy1, dx2, dy2, dx3]
							pendingOp = "vhcurveto"
					else:
						if  pendingOp != kNoOp:
							newT2List.append([arglist, pendingOp])
						arglist = [dy1, dx2, dy2, dx3]
						pendingOp = sequenceOp = "vhcurveto"

				elif dx3 == 0: # - dy1 dx2 dy2 - dy3 vvcurveto
					if pendingOp not in ["vvcurveto", kNoOp]:
						newT2List.append([arglist, pendingOp])
						arglist = []
					arglist.extend([dy1, dx2, dy2, dy3 ])
					sequenceOp = "vvcurveto"
					if len(arglist) >= kStackLimit:
						newT2List.append([arglist[:-4], pendingOp])
						arglist = [dy1, dx2, dy2, dy3]
						pendingOp =  sequenceOp
					else:
						pendingOp = sequenceOp

				else:  # - dy1 dx2 dy2 dx3 dy3 vhcurveto (odd number of args, can't concatenate any more ops.)
					if  (pendingOp in ["vhcurveto", "hvcurveto"]) and (sequenceOp ==  "hvcurveto"):
						arglist.extend([dy1, dx2, dy2, dx3, dy3])
						if len(arglist) >= kStackLimit:
							newT2List.append([arglist[:-5], pendingOp])
							arglist = [dy1, dx2, dy2, dx3, dy3]
							pendingOp = "vhcurveto"
					else:
						if  pendingOp != kNoOp:
							newT2List.append([arglist, pendingOp])
						arglist = [dy1, dx2, dy2, dx3, dy3]
						pendingOp = "vhcurveto"
					newT2List.append([arglist, pendingOp])
					arglist = []
					pendingOp = sequenceOp = kNoOp

			elif dy1 == 0:
				if dx3 == 0: # dx1 - dx2 dy2 - dy3 hvcurveto
					if  (pendingOp in ["vhcurveto", "hvcurveto"]) and (sequenceOp ==  "vhcurveto"):
						arglist.extend([dx1, dx2, dy2, dy3])
						sequenceOp = "hvcurveto"
						if len(arglist) >= kStackLimit:
							newT2List.append([arglist[:-4], pendingOp])
							arglist = [dx1, dx2, dy2, dy3]
							pendingOp = "hvcurveto"
					else:
						if  pendingOp != kNoOp:
							newT2List.append([arglist, pendingOp])
						arglist = [dx1, dx2, dy2, dy3]
						pendingOp = sequenceOp = "hvcurveto"

				elif dy3 == 0: # dx1 - dx2 dy2 dx3 - hhcurveto
					if pendingOp not in ["hhcurveto", kNoOp]:
						newT2List.append([arglist, pendingOp])
						arglist = []
					arglist.extend([dx1, dx2, dy2, dx3 ])
					sequenceOp = "hhcurveto"
					if len(arglist) >= kStackLimit:
						newT2List.append([arglist[:-4], pendingOp]) # XXX Problem. Was vvcurveto
						arglist = [dx1, dx2, dy2, dx3]
						pendingOp = sequenceOp
					else:
						pendingOp = sequenceOp

				else:  # dx1 - dx2 dy2 dy3 dx3 hvcurveto (odd number of args, can't concatenate any more ops.)
					if  (pendingOp in ["vhcurveto", "hvcurveto"]) and (sequenceOp ==  "vhcurveto"):
						arglist.extend( [dx1, dx2, dy2, dy3, dx3])
						if len(arglist) >= kStackLimit:
							newT2List.append([arglist[:-5], pendingOp])
							arglist =  [dx1, dx2, dy2, dy3, dx3]
							pendingOp = "hvcurveto"
					else:
						if  pendingOp != kNoOp:
							newT2List.append([arglist, pendingOp])
						arglist = [dx1, dx2, dy2, dy3, dx3]
						pendingOp = "hvcurveto"
					newT2List.append([arglist, pendingOp])
					arglist = []
					pendingOp = sequenceOp = kNoOp

			elif dx3 == 0: #  dx1 dy1 dx2 dy2 - dy3 vvcurveto (odd args)
					if pendingOp != kNoOp:
						newT2List.append([arglist, pendingOp])
						arglist = []
					arglist = [dx1, dy1, dx2, dy2, dy3]
					pendingOp = "vvcurveto"
					newT2List.append([arglist, pendingOp])
					arglist = []
					pendingOp = sequenceOp = kNoOp

			elif dy3 == 0: #  dx1 dy1 dx2 dy2 dx3 - hhcurveto (odd args)
					if pendingOp != kNoOp:
						newT2List.append([arglist, pendingOp])
						arglist = []
					arglist = [dy1, dx1, dx2, dy2, dx3] # note arg order swap
					pendingOp = "hhcurveto"
					newT2List.append([arglist, pendingOp])
					arglist = []
					pendingOp = sequenceOp = kNoOp

			else:
				if pendingOp == "rlineto":
					arglist.extend(args)
					if len(arglist) >= kStackLimit:
						newT2List.append([arglist[:-len(args)], pendingOp])
						arglist = args
						pendingOp = sequenceOp = op
					else:
						newT2List.append([arglist, "rlinecurve"])
						arglist = []
						pendingOp = sequenceOp = kNoOp

				else:
					if pendingOp not in [kNoOp, "rrcurveto"]:
						newT2List.append([arglist, pendingOp])
						arglist = []
					arglist.extend(args)
					if len(arglist) >= kStackLimit:
						newT2List.append([arglist[:-len(args)], pendingOp])
						arglist = args
					pendingOp = sequenceOp = op

		elif op ==  "flex":
			dx1 = args[0]
			dy1 = args[1]
			dx2 = args[2]
			dy2 = args[3]
			dx3 = args[4]
			dy3 = args[5]
			dx4 = args[6]
			dy4 = args[7]
			dx5 = args[8]
			dy5 = args[9]
			dx6 = args[10]
			dy6 = args[11]
			if pendingOp != kNoOp:
				newT2List.append([arglist, pendingOp])
				arglist = []
			noFlex = 1
			noDY = 1
			if (dy3 == 0 == dy4):
				if (dy1 == dy6 == 0) and (dy2 == -dy5):
						newT2List.append([[dx1, dx2, dy2, dx3, dx4, dx5, dx6], "hflex"]) # the device pixel threshold is always 50 , when coming back from AC.
						noFlex = 0
				else:
					dy = dy1 + dy2 + dy3 + dy4 + dy5 + dy6
					noDY = 0
					if dy == 0:
						newT2List.append([[dx1, dy1, dx2, dy2, dx3, dx4, dx5, dy5, dx6], "hflex1"])
						noFlex = 0

			if noFlex:
				if 0:
					dx = dx1 + dx2 + dx3 + dx4 + dx5
					dy = dy1 + dy2 + dy3 + dy4 + dy5

					if ((dy + dy6) == 0) or ((dx+dx6) == 0):
						if abs(dx) > abs(dy):
							lastArg = dx6
						else:
							lastArg = dy6

						newT2List.append([args[:10] + [lastArg], "flex1"])
						newLastArg = lastArg
					else:
						newT2List.append([args, "flex"])
				else:
					newT2List.append([args, "flex"])

			arglist = []
			pendingOp = sequenceOp = kNoOp


		else:
			if pendingOp != kNoOp:
				newT2List.append([arglist, pendingOp])
			newT2List.append([args, op])
			arglist = []
			pendingOp = sequenceOp = kNoOp

	if pendingOp != kNoOp:
		newT2List.append([arglist, pendingOp])


	return newT2List

def needsDecryption(bezDataBuffer):
	lenBuf = len(bezDataBuffer)
	if lenBuf > 100:
		lenBuf = 100
	i = 0
	while i < lenBuf:
		ch = bezDataBuffer[i]
		i +=1
		if not (ch.isspace() or ch.isdigit()  or (ch in "ABCDEFabcdef") ):
			return 0
	return 1

LEN_IV  = 4 #/* Length of initial random byte sequence */
def  bezDecrypt(bezDataBuffer):
	r = 11586
	i = 0 # input buffer byte position index
	lenBuffer = len(bezDataBuffer)
	byteCnt = 0 # output buffer byte count.
	newBuffer = ""
	while 1:
		cipher = 0 # restricted to int
		plain = 0 # restricted to int
		j = 2 # used to combine two successive bytes

		# process next two bytes, skipping whitespace.
		while j > 0:
			j -=1
			try:
				while  bezDataBuffer[i].isspace():
					i +=1
				ch = bezDataBuffer[i]
			except IndexError:
				return newBuffer

			if not ch.islower():
				ch = ch.lower()
			if ch.isdigit():
				ch = byteord(ch) - byteord('0')
			else:
				ch = byteord(ch) - byteord('a') + 10
			cipher = (cipher << 4) & 0xFFFF
			cipher = cipher | ch
			i += 1

		plain = cipher ^ (r >> 8)
		r = (cipher + r) * 902381661 + 341529579
		if r  > 0xFFFF:
			r = r & 0xFFFF
		byteCnt +=1
		if (byteCnt > LEN_IV):
			newBuffer += bytechr(plain)
		if i >= lenBuffer:
			break

	return newBuffer


kHintArgsNoOverlap = 0
kHintArgsOverLap = 1
kHintArgsMatch = 2

def checkStem3ArgsOverlap(argList, hintList):
	# status == 0 -> no overlap
	# status == 1 -> arg are the same
	# status = 2 -> args  overlap, and are not the same
	status = kHintArgsNoOverlap
	for x0,x1 in argList:
		x1 = x0 + x1
		for y0,y1 in hintList:
			y1 = y0 +y1
			if ( x0 == y0):
				if (x1 == y1):
					status = kHintArgsMatch
				else:
					return kHintArgsOverLap
			elif (x1 == y1):
				return kHintArgsOverLap
			else:
				if (x0 > y0) and (x0 < y1):
					return kHintArgsOverLap
				if (x1 > y0) and (x1 < y1):
					return kHintArgsOverLap
	return status


def buildControlMaskList(hStem3List, vStem3List):
	""" The deal is that a char string will use either counter hints, or stem 3 hints, but
	 not both. We examine all the arglists. If any are not a multiple of 3, then we use all the arglists as is as the args to a counter hint.
	 If all are a multiple of 3, then we divide them up into triplets, adn add a separate conter mask for each unique arg set.
	"""

	vControlMask =  HintMask(0)
	hControlMask = vControlMask
	controlMaskList = [hControlMask]
	for argList in hStem3List:
		for mask in controlMaskList:
			overlapStatus = kHintArgsNoOverlap
			if not mask.hList:
				mask.hList.extend(argList)
				overlapStatus = kHintArgsMatch
				break
			overlapStatus = checkStem3ArgsOverlap(argList, mask.hList)
			if overlapStatus == kHintArgsMatch: # The args match args in this control mask.
				break
		if overlapStatus != kHintArgsMatch:
			mask = HintMask(0)
			controlMaskList.append(mask)
			mask.hList.extend(argList)

	for argList in vStem3List:
		for mask in controlMaskList:
			overlapStatus = kHintArgsNoOverlap
			if not mask.vList:
				mask.vList.extend(argList)
				overlapStatus = kHintArgsMatch
				break
			overlapStatus = checkStem3ArgsOverlap(argList, mask.vList)
			if overlapStatus == kHintArgsMatch: # The args match args in this control mask.
				break
		if overlapStatus != kHintArgsMatch:
			mask = HintMask(0)
			controlMaskList.append(mask)
			mask.vList.extend(argList)

	return controlMaskList

def makeRelativeCTArgs(argList, curX, curY):
	newCurX = argList[4]
	newCurY = argList[5]
	argList[5] -= argList[3]
	argList[4] -= argList[2]

	argList[3] -= argList[1]
	argList[2] -= argList[0]

	argList[0] -= curX
	argList[1] -= curY
	return argList, newCurX, newCurY

def convertBezToT2(bezString):
	# convert bez data to a T2 outline program, a list of operator tokens.
	#
	# Convert all bez ops to simplest T2 equivalent
	# Add all hints to vertical and horizontal hint lists as encountered; insert a HintMask class whenever a
	# new set of hints is encountered
	# after all operators have been processed, convert HintMask items into hintmask ops and hintmask bytes
	# add all hints as prefix
	# review operator list to optimize T2 operators.

	bezString = re.sub(r"%.+?\n", "", bezString) # supress comments
	bezList = re.findall(r"(\S+)", bezString)
	if not bezList:
		return ""
	hhints = []
	vhints = []
	hintMask = HintMask(0) # Always assume a hint mask until proven otherwise.
	hintMaskList = [hintMask]
	vStem3Args = []
	hStem3Args = []
	vStem3List = []
	hStem3List = []
	argList = []
	t2List = []

	lastPathOp = None
	curX = 0
	curY = 0
	for token in bezList:
		try:
			val1 = round(float(token),2)
			try:
				val2 = int(token)
				if int(val1) == val2:
					argList.append(val2)
				else:
					argList.append("% 100 div" % (str(int(val1*100))))
			except ValueError:
				argList.append(val1)
			continue
		except ValueError:
			pass

		if token == "newcolors":
			lastPathOp = token
			pass
		elif token in ["beginsubr", "endsubr"]:
			lastPathOp = token
			pass
		elif token in ["snc"]:
			lastPathOp = token
			hintMask = HintMask(len(t2List)) # The index into the t2list is kept so we can quickly find them later.
			t2List.append([hintMask])
			hintMaskList.append(hintMask)
		elif token in ["enc"]:
			lastPathOp = token
			pass
		elif token == "div":
			# i specifically do NOT set lastPathOp for this.
			value = argList[-2]/float(argList[-1])
			argList[-2:] =[value]
		elif token == "rb":
			lastPathOp = token
			try:
				i = hhints.index(argList)
			except ValueError:
				i = len(hhints)
				hhints.append(argList)
			if hintMask:
				if hhints[i] not in hintMask.hList:
					hintMask.hList.append(hhints[i])
			argList = []
		elif token == "ry":
			lastPathOp = token
			try:
				i = vhints.index(argList)
			except ValueError:
				i = len(vhints)
				vhints.append(argList)
			if hintMask:
				if vhints[i] not in hintMask.vList:
					hintMask.vList.append(vhints[i])
			argList = []

		elif token == "rm": # vstem3 hints are vhints
			try:
				i = vhints.index(argList)
			except ValueError:
				i = len(vhints)
				vhints.append(argList)
			if hintMask:
				if vhints[i] not in hintMask.vList:
					hintMask.vList.append(vhints[i])

			if (lastPathOp != token) and vStem3Args:
				# first rm, must be start of a new vstem3
				# if we already have a set of vstems in vStem3Args, save them,
				# and then cleae the vStem3Args so we can add the new set.
				vStem3List.append(vStem3Args)
				vStem3Args = []

			vStem3Args.append(argList)
			argList = []
			lastPathOp = token
		elif token == "rv": # hstem3 are hhints
			try:
				i = hhints.index(argList)
			except ValueError:
				i = len(hhints)
				hhints.append(argList)
			if hintMask:
				if hhints[i] not in hintMask.hList:
					hintMask.hList.append(hhints[i])

			if (lastPathOp != token) and hStem3Args:
				# first rv, must be start of a new h countermask
				hStem3List.append(hStem3Args)
				hStem3Args = []

			hStem3Args.append(argList)
			argList = []
			lastPathOp = token
		elif token == "preflx1":
			# the preflx1/  preflx2 sequence provides the same i as the flex sequence; the difference is that the
			#  preflx1/preflx2 sequence provides the argument values needed for building a Type1 string
			# while the flex sequence is simply the 6 rcurveto points. Both sequences are always provided.
			lastPathOp = token
			argList = []
		elif token in ["preflx2", "preflx2a"]:
			lastPathOp = token
			del t2List[-1]
			argList = []
		elif token == "flx":
			lastPathOp = token
			argList = argList[:12]
			t2List.append([argList + [50], "flex"])
			argList = []
		elif token == "flxa":
			lastPathOp = token
			argList1, curX, curY = makeRelativeCTArgs(argList[:6], curX, curY)
			argList2, curX, curY = makeRelativeCTArgs(argList[6:], curX, curY)
			argList = argList1 + argList2
			t2List.append([argList[:12] + [50], "flex"])
			argList = []
		elif token == "sc":
			lastPathOp = token
			pass
		else:
			if token[-2:] in ["mt", "dt", "ct", "cv"]:
				lastPathOp = token
			t2Op = bezToT2.get(token,None)
			if token in ["mt", "dt"]:
				newList = [argList[0] - curX, argList[1]- curY]
				curX = argList[0]
				curY = argList[1]
				argList = newList
			elif token in ["ct", "cv"]:
				argList, curX, curY = makeRelativeCTArgs(argList, curX, curY)
			if t2Op:
				t2List.append([argList, t2Op])
			elif t2Op == None:
				print("Unhandled operation", argList, token)
				raise KeyError
			argList = []


	# add hints, if any
	# Must be done at the end of op processing to make sure we have seen all the
	# hints in the bez string.
	# Note that the hintmask are identified in the t2List by an index into the list
	# be careful to NOT change the t2List length until the hintmasks have been converted.
	numHintMasks = len(hintMaskList)
	needHintMasks = numHintMasks > 1

	if vStem3Args:
		vStem3List.append(vStem3Args)
	if hStem3Args :
		hStem3List.append(hStem3Args)


	t2Program = []
	hhints.sort()
	vhints.sort()
	numHHints = len(hhints)
	numVHints =  len(vhints)
	hintLimit = (kStackLimit - 2) // 2
	if numHHints >=hintLimit:
		hhints = hhints[:hintLimit]
		numHHints = hintLimit
	if numVHints >=hintLimit:
		vhints = vhints[:hintLimit]
		numVHints = hintLimit
	if hhints:
		isH = 1
		t2Program = makeHintList(hhints, needHintMasks, isH)
	if vhints:
		isH = 0
		t2Program += makeHintList(vhints, needHintMasks, isH)

	if vStem3List or hStem3List:
		controlMaskList = buildControlMaskList(hStem3List, vStem3List)
		for cMask in controlMaskList:
			hBytes = cMask.maskByte(hhints, vhints)
			t2Program.extend(["cntrmask", hBytes])

	if needHintMasks:
		# If there is not a hintsub before any drawing operators, then
		# add an initial first hint mask to the t2Program.
		if hintMaskList[1].listPos != 0:
			hBytes = hintMaskList[0].maskByte(hhints, vhints)
			t2Program.extend(["hintmask", hBytes])

		# Convert the rest of the hint masks to a hintmask op and hintmask bytes.
		for hintMask in hintMaskList[1:]:
			pos = hintMask.listPos
			t2List[pos] = [["hintmask"], hintMask.maskByte(hhints, vhints)]

	t2List = optimizeT2Program(t2List)

	for entry in t2List:
		try:
			t2Program.extend(entry[0])
			t2Program.append(entry[1])
		except:
			print("Failed to extend t2Program with entry", entry)
			raise KeyError

	return t2Program


class CFFFontData:
	def __init__(self, ttFont, inputPath, outFilePath, fontType, logMsgCB):
		self.ttFont = ttFont
		self.inputPath = inputPath
		if (outFilePath == None):
			outFilePath = inputPath
		self.outFilePath = outFilePath
		self.fontType = fontType
		self.logMsg = logMsgCB
		try:
			self.cffTable = ttFont["CFF "]
			topDict =  self.cffTable.cff.topDictIndex[0]
		except KeyError:
			raise focusFontError("Error: font is not a CFF font <%s>." % fontFileName)

		#    for identifier in glyph-list:
		# 	Get charstring.
		self.topDict = topDict
		self.charStrings = topDict.CharStrings
		self.charStringIndex = self.charStrings.charStringsIndex
		self.allowDecimalCoords = False

	def getGlyphList(self):
		fontGlyphList = self.ttFont.getGlyphOrder()
		return fontGlyphList

	def getUnitsPerEm(self):
		unitsPerEm = "1000"
		if hasattr(self.topDict, "FontMatrix"):
			matrix = self.topDict.FontMatrix
			unitsPerEm = "%s" % (int(round(1.0/matrix[0])))
		return unitsPerEm

	def getPSName(self):
		psName = self.cffTable.cff.fontNames[0]
		return psName

	def convertToBez(self, glyphName, removeHints, beVerbose, doAll=False):
		hasHints = 0
		t2Wdth = None
		gid = self.charStrings.charStrings[glyphName]
		t2CharString = self.charStringIndex[gid]
		try:
			bezString, hasHints, t2Wdth = convertT2GlyphToBez(t2CharString, removeHints, self.allowDecimalCoords)
			# Note: the glyph name is important, as it is used by autohintexe for various heurisitics, including [hv]stem3 derivation.
			bezString  = (r"%%%s%s " % (glyphName, os.linesep)) + bezString
		except SEACError:
			if not beVerbose:
				dotCount = 0
				self.logMsg("") # end series of "."
				self.logMsg("Checking %s -- ," % (glyphName)) # output message when SEAC glyph is found
			self.logMsg("Skipping %s: can't process SEAC composite glyphs." % (glyphName))
			bezString = None
		return bezString, t2Wdth, hasHints

	def updateFromBez(self, bezData, glyphName, width, beVerbose):
		t2Program = [width] + convertBezToT2(bezData)
		if t2Program:
			gid = self.charStrings.charStrings[glyphName]
			t2CharString = self.charStringIndex[gid]
			t2CharString.program = t2Program
		else:
			if not beVerbose:
				dotCount = 0
				self.logMsg("") # end series of "."
				self.logMsg("Checking %s -- ," % (aliasName(name)))
			self.logMsg("Skipping %s: error in processing fixed outline." % (aliasName(name)))

	def saveChanges(self):
		ttFont = self.ttFont
		fontType = self.fontType
		inputPath = self.inputPath
		outFilePath = self.outFilePath

		overwriteOriginal = 0
		if inputPath == outFilePath:
			overwriteOriginal = 1
		tempPath = "{}.temp.ac".format(inputPath)

		if fontType == 0: # OTF
			if overwriteOriginal:
				ttFont.save(tempPath)
				ttFont.close()
				if os.path.exists(inputPath):
					try:
						os.remove(inputPath)
						os.rename(tempPath, inputPath)
					except (OSError, IOError):
						self.logMsg( "\t%s" %(traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1])[-1]))
						self.logMsg("Error: could not overwrite original font file path '%s'. Hinted font file path is '%s'." % (inputPath, tempPath))
			else:
				ttFont.save(outFilePath)
				ttFont.close()

		else:
			data = ttFont["CFF "].compile(ttFont)
			if fontType == 1: # CFF
				if overwriteOriginal:
					tf = open(inputPath, "wb")
					tf.write(data)
					tf.close()
				else:
					tf = open(outFilePath, "wb")
					tf.write(data)
					tf.close()

			elif  fontType == 2: # PS.
				tf = open(tempPath, "wb")
				tf.write(data)
				tf.close()
				finalPath = outFilePath
				command="tx  -t1 -std \"%s\" \"%s\" 2>&1" % (tempPath, outFilePath)
				report = fdkutils.runShellCmd(command)
				self.logMsg(report)
				if "fatal" in report:
					raise IOError("Failed to convert hinted font temp file with tx %s. Maybe target font font file '%s' is set to read-only. %s" % (tempPath, outFilePath, report))

			if os.path.exists(tempPath):
				os.remove(tempPath)
	def close(self):
		self.ttFont.close()

	def getGlyphID(self, name):
		gid = self.ttFont.getGlyphID(name)
		return gid

	def isCID(self):
		isCID = hasattr(self.topDict, "FDSelect")
		return isCID

	def getFontInfo(self, fontPSName, inputPath, allow_no_blues, noFlex, vCounterGlyphs, hCounterGlyphs,fdIndex = 0):
		# The AC library needs the global font hint zones and standard stem widths. Format them
		# into a single  text string.
		# The text format is arbitrary, inherited from very old software, but there is no real need to change it.
		pTopDict = self.topDict
		if  hasattr(pTopDict, "FDArray"):
			pDict = pTopDict.FDArray[fdIndex]
		else:
			pDict = pTopDict
		privateDict = pDict.Private

		fdDict = convertfonttocid.FDDict()
		if  hasattr(privateDict, "LanguageGroup"):
			fdDict.LanguageGroup = privateDict.LanguageGroup
		else:
			fdDict.LanguageGroup = "0"

		if  hasattr(pDict, "FontMatrix"):
			fdDict.FontMatrix = pDict.FontMatrix
		else:
			fdDict.FontMatrix = pTopDict.FontMatrix
		upm = int(1/fdDict.FontMatrix[0])
		fdDict.OrigEmSqUnits = str(upm)

		if  hasattr(pTopDict, "FontName"):
			fdDict.FontName = pDict.FontName # FontName
		else:
			fdDict.FontName = fontPSName

		low =  min( -upm*0.25, pTopDict.FontBBox[1] - 200)
		high =  max ( upm*1.25, pTopDict.FontBBox[3] + 200)
		# Make a set of inactive alignment zones: zones outside of the font bbox so as not to affect hinting.
		# Used when src font has no BlueValues or has invalid BlueValues. Some fonts have bad BBOx values, so
		# I don't let this be smaller than -upm*0.25, upm*1.25.
		inactiveAlignmentValues = [low, low, high, high]
		if hasattr(privateDict, "BlueValues"):
			blueValues = privateDict.BlueValues[:]
			numBlueValues = len(privateDict.BlueValues)
			blueValues.sort()
			if numBlueValues < 4:
				if allow_no_blues:
					blueValues = inactiveAlignmentValues
					numBlueValues = len(blueValues)
				else:
					raise	ACFontError("Error: font must have at least four values in its BlueValues array for AC to work!")
		else:
			if allow_no_blues:
				blueValues = inactiveAlignmentValues
				numBlueValues = len(blueValues)
			else:
				raise	ACFontError("Error: font has no BlueValues array!")

		# The first pair only is a bottom zone,, where the first value is the overshoot position;
		# the rest are top zones, and second value of the pair is the overshoot position.
		blueValues[0] = blueValues[0] - blueValues[1]
		for i in range(3, numBlueValues,2):
			blueValues[i] = blueValues[i] - blueValues[i-1]

		blueValues = [str(v) for v in blueValues]
		numBlueValues = min(numBlueValues, len(convertfonttocid.kBlueValueKeys))
		for i in range(numBlueValues):
			key = convertfonttocid.kBlueValueKeys[i]
			value = blueValues[i]
			exec("fdDict.%s = %s" % (key, value))

		#print numBlueValues
		#for i in range(0, len(fontinfo),2):
		#	print fontinfo[i], fontinfo[i+1]

		if hasattr(privateDict, "OtherBlues"):
			# For all OtherBlues, the pairs are bottom zones, and the first value of each pair is the overshoot position.
			i = 0
			numBlueValues = len(privateDict.OtherBlues)
			blueValues = privateDict.OtherBlues[:]
			blueValues.sort()
			for i in range(0, numBlueValues,2):
				blueValues[i] = blueValues[i] - blueValues[i+1]
			blueValues = [str(val) for val in blueValues]
			numBlueValues = min(numBlueValues, len(convertfonttocid.kOtherBlueValueKeys))
			for i in range(numBlueValues):
				key = convertfonttocid.kOtherBlueValueKeys[i]
				value = blueValues[i]
				exec("fdDict.%s = %s" % (key, value))

		if hasattr(privateDict, "StemSnapV"):
			vstems = privateDict.StemSnapV
		elif hasattr(privateDict, "StdVW"):
			vstems = [privateDict.StdVW]
		else:
			if allow_no_blues:
				vstems =  [upm] # dummy value. Needs to be larger than any hint will likely be,
				# as the autohint program strips out any hint wider than twice the largest global stem width.
			else:
				raise	ACFontError("Error: font has neither StemSnapV nor StdVW!")
		vstems.sort()
		if (len(vstems) == 0) or ((len(vstems) == 1) and (vstems[0] < 1) ):
			vstems =  [upm] # dummy value that will allow PyAC to run
			self.logMsg("Warning: There is no value or 0 value for DominantV.")
		vstems = repr(vstems)
		fdDict.DominantV = vstems

		if hasattr(privateDict, "StemSnapH"):
			hstems = privateDict.StemSnapH
		elif hasattr(privateDict, "StdHW"):
			hstems = [privateDict.StdHW]
		else:
			if allow_no_blues:
				hstems = [upm] # dummy value. Needs to be larger than any hint will likely be,
				# as the autohint program strips out any hint wider than twice the largest global stem width.
			else:
				raise	ACFontError("Error: font has neither StemSnapH nor StdHW!")
		hstems.sort()
		if (len(hstems) == 0) or ((len(hstems) == 1) and (hstems[0] < 1) ):
			hstems =  [upm] # dummy value that will allow PyAC to run
			self.logMsg("Warning: There is no value or 0 value for DominantH.")
		hstems = repr(hstems)
		fdDict.DominantH = hstems

		if noFlex:
			fdDict.FlexOK = "false"
		else:
			fdDict.FlexOK = "true"

		# Add candidate lists for counter hints, if any.
		if vCounterGlyphs:
			temp = " ".join(vCounterGlyphs)
			fdDict.VCounterChars = "( %s )" % (temp)
		if hCounterGlyphs:
			temp = " ".join(hCounterGlyphs)
			fdDict.HCounterChars = "( %s )" % (temp)

		if  hasattr(privateDict, "BlueFuzz"):
			fdDict.BlueFuzz = privateDict.BlueFuzz
		else:
			fdDict.BlueFuzz = 1

		return fdDict

	def getfdIndex(self, gid):
		fdIndex = self.topDict.FDSelect[gid]
		return fdIndex

	def getfdInfo(self,  fontPSName, inputPath, allow_no_blues, noFlex, vCounterGlyphs, hCounterGlyphs, glyphList, fdIndex=0):
		topDict = self.topDict
		fontDictList = []
		fdGlyphDict = None

		# Get the default fontinfo from the font's top dict.
		fdDict = self.getFontInfo(fontPSName, inputPath, allow_no_blues, noFlex, vCounterGlyphs, hCounterGlyphs, fdIndex)
		fontDictList.append(fdDict)

		# Check the fontinfo file, and add any other font dicts
		srcFontInfo = os.path.dirname(inputPath)
		srcFontInfo = os.path.join(srcFontInfo, "fontinfo")
		if os.path.exists(srcFontInfo):
			fi = open(srcFontInfo, "r")
			fontInfoData = fi.read()
			fi.close()
			fontInfoData = re.sub(r"#[^\r\n]+", "", fontInfoData)
		else:
			return  fdGlyphDict, fontDictList

		if "FDDict" in fontInfoData:
			maxY = topDict.FontBBox[3]
			minY = topDict.FontBBox[1]
			blueFuzz = convertfonttocid.getBlueFuzz(inputPath)
			fdGlyphDict, fontDictList, finalFDict = convertfonttocid.parseFontInfoFile(fontDictList, fontInfoData, glyphList, maxY, minY, fontPSName, blueFuzz)
			if finalFDict == None:
				# If a font dict was not explicitly specified for the output font, use the first user-specified font dict.
				convertfonttocid.mergeFDDicts( fontDictList[1:], topDict.Private )
			else:
				convertfonttocid.mergeFDDicts( [finalFDict], topDict.Private )
		return fdGlyphDict, fontDictList



from __future__ import print_function, division, absolute_import

import pytest
import subprocess32 as subprocess

from runner import main as runner
from differ import main as differ, SPLIT_MARKER
from test_utils import get_expected_path

TOOL = 'spot'
CMD = ['-t', TOOL]


# -----
# Tests
# -----

@pytest.mark.parametrize('arg', ['-h', '-u'])
def test_exit_known_option(arg):
    assert subprocess.call([TOOL, arg]) == 0


@pytest.mark.parametrize('arg', ['-z', '-foo'])
def test_exit_unknown_option(arg):
    assert subprocess.call([TOOL, arg]) == 1


@pytest.mark.parametrize('args, exp_filename', [
    (['T'], 'list_tables.txt'),
    (['F'], 'list_features.txt'),
    (['G'], 'glyph_proof.ps'),  # same as dump_CFF_=7
    (['G', 'd'], 'glyph_proof_no_header.ps'),
    (['G', 'd', 'br'], 'glyph_proof_no_header_multipage.ps'),
    (['Proof'], 'proof_all_features.ps'),
    (['t', '_BASE=4'], 'dump_BASE=4.txt'),
    (['t', '_BASE=5'], 'dump_BASE=5.txt'),
    (['t', '_GDEF=7'], 'dump_GDEF=7.txt'),
    (['t', '_GPOS=5'], 'dump_GPOS=5.txt'),
    (['t', '_GPOS=6'], 'dump_GPOS=6.txt'),
    (['t', '_GPOS=7'], 'dump_GPOS=7.txt'),
    (['t', '_GPOS=8'], 'dump_GPOS=8.ps'),
    (['t', '_GSUB=4'], 'dump_GSUB=4.txt'),
    (['t', '_GSUB=5'], 'dump_GSUB=5.txt'),
    (['t', '_GSUB=7'], 'dump_GSUB=7.txt'),
    (['t', '_GSUB=8'], 'dump_GSUB=8.ps'),
    (['t', '_cmap=5'], 'dump_cmap=5.txt'),
    (['t', '_cmap=6'], 'dump_cmap=6.txt'),
    (['t', '_cmap=7'], 'dump_cmap=7.txt'),
    (['t', '_cmap=8'], 'dump_cmap=8.txt'),
    (['t', '_cmap=9'], 'dump_cmap=9.ps'),
    (['t', '_cmap=10'], 'dump_cmap=10.ps'),
    (['t', '_cmap=11'], 'dump_cmap=11.txt'),
    (['t', '_hmtx=5'], 'dump_hmtx=5.txt'),
    (['t', '_hmtx=6'], 'dump_hmtx=6.txt'),
    (['t', '_hmtx=7'], 'dump_hmtx=7.txt'),
    (['t', '_hmtx=8'], 'dump_hmtx=8.txt'),
    (['t', '_name=2'], 'dump_name=2.txt'),
    (['t', '_name=3'], 'dump_name=3.txt'),
    (['t', '_name=4'], 'dump_name=4.txt'),
    (['t', '_CFF_=6'], 'dump_CFF_=6.ps'),
    (['t', '_CFF_=6', 'a'], 'dump_CFF_=6a.ps'),
    (['t', '_CFF_=6', 'c'], 'dump_CFF_=6c.ps'),
    (['t', '_CFF_=6', 'R'], 'dump_CFF_=6R.ps'),
    (['t', '_CFF_=6', 'g2,5-7'], 'dump_CFF_=6g.ps'),
    (['t', '_CFF_=6', 's1.5,0.5'], 'dump_CFF_=6s.ps'),
    (['t', '_CFF_=6', 'b-10,-10,200,500', 'g4'], 'dump_CFF_=6b.ps'),
    (['t', '_CFF_=7'], 'dump_CFF_=7.ps'),
    (['t', '_CFF_=7', 'g2,5-7'], 'dump_CFF_=7g.ps'),
    (['t', '_CFF_=8'], 'dump_CFF_=8.ps'),
    (['t', '_CFF_=8', 'g2,5-7'], 'dump_CFF_=8g.ps'),
])
def test_options(args, exp_filename):
    if 'CFF_=7' in exp_filename or exp_filename == 'glyph_proof.ps':
        skip = ['0 743.4 moveto (CFF:']
    elif 'CFF_=6' in exp_filename:
        skip = ['318 764 moveto (' + SPLIT_MARKER + '72 750 moveto (']
    else:
        skip = []
    if skip:
        skip.insert(0, '-s')
    actual_path = runner(CMD + ['-s', '-f', 'black.otf', '-o'] + args)
    expected_path = get_expected_path(exp_filename)
    assert differ([expected_path, actual_path] + skip)


@pytest.mark.parametrize('table', ['GSUB', 'GPOS'])
def test_feature_tags_bug691(table):
    exp_filename = 'SourceCodePro-Regular_{}.ps'.format(table)
    actual_path = runner(CMD + ['-s', '-f', 'SourceCodePro-Regular.otf',
                                '-o', 't', '_{}=8'.format(table)])
    expected_path = get_expected_path(exp_filename)
    assert differ([expected_path, actual_path])


@pytest.mark.parametrize('font_format', ['otf', 'ttf'])
def test_long_glyph_name_bug373(font_format):
    file_name = 'long_glyph_name.' + font_format
    actual_path = runner(CMD + ['-s', '-o', 't', '_GSUB=7', '-f', file_name])
    expected_path = get_expected_path('bug373_{}.txt'.format(font_format))
    assert differ([expected_path, actual_path])


def test_buffer_overrun_bug465():
    """ Fix bug where a fixed length string buffer was overrun. The test
    font was built by building an otf from makeotf_data/input/t1pfa.pfa,
    dumping this to ttx, and then hand-editing the ttx file to have
    empty charstrings, and a ContextPos lookup type 7, format 2.
    Hand-editing the ttx was needed as makeotf does not build this
    lookup type, nor this post rule format. This test validates that the
    current spot code correctly reports the test font GPOS table."""
    file_name = "bug465/bug465.otf"
    actual_path = runner(CMD + ['-s', '-o', 't', '_GPOS=7', '-f', file_name])
    expected_path = get_expected_path('bug465_otf.txt')
    assert differ([expected_path, actual_path])



# Copyright 2017 Adobe. All rights reserved.

from __future__ import print_function, division, absolute_import

import os
import shutil
import sys
import tempfile

from fontTools.misc.py23 import tounicode

import defcon
from mutatorMath.ufo.document import DesignSpaceDocumentReader

from afdko.fdkutils import runShellCmd

try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET

XMLElement = ET.Element
xmlToString = ET.tostring

kTempUFOExt = ".temp.ufo"
kTempDSExt = ".temp.designspace"
kFeaturesFile = "features.fea"

__usage__ = """
buildmasterotfs.py  1.8.2 Jan 29 2019
Build master source OpenType/CFF fonts from a Superpolator design space file
and the UFO master source fonts.

python buildmasterotfs.py -h
python buildmasterotfs.py -u
python buildmasterotfs.py  <path to design space file>
"""

__help__ = __usage__ + """
Options:
--mkot   Allows passing a comma-separated set of options to makeotf

The script makes a number of assumptions.
1) all the master source fonts are blend compatible in all their data.
2) Each master source font is in a separate directory from the other master
source fonts, and each of these directories provides the default meta data
files expected by the 'makeotf' command:
- features/features.fea
- font.ufo
- FontMenuNameDB within the directory or no more than 4 parent directories up.
- GlyphAliasAndOrderDB within the directory or no more than 4 parent
  directories up.
3) The Adobe AFDKO is installed: the script calls 'tx' and 'makeotf'.
4) <source> element for the default font in the design space file has
the child element <info copy="1">; this is used to identify which master is
the default font.

The script will build a master source OpenType/CFF font for each master source
UFO font, using the same file name, but with the extension '.otf'.

UFO masters with partial glyphsets are supported.
"""


def compatibilizePaths(otfPath):
    tempPathCFF = otfPath + ".temp.cff"
    command = "tx -cff +b -std -no_opt \"%s\" \"%s\" 2>&1" % (otfPath,
                                                              tempPathCFF)
    report = runShellCmd(command)
    if "fatal" in str(report):
        print(report)
        sys.exit(1)

    command = "sfntedit -a \"CFF \"=\"%s\" \"%s\" 2>&1" % (tempPathCFF,
                                                           otfPath)
    report = runShellCmd(command)
    if "FATAL" in str(report):
        print(report)
        sys.exit(1)

    os.remove(tempPathCFF)


def _determine_which_masters_to_generate(ds_path):
    """'ds_path' is the path to a designspace file.
       Returns a list of integers representing the indexes
       of the temp masters to generate.
    """
    master_list = ET.parse(ds_path).getroot().find('sources').findall('source')

    # Make a set of the glyphsets of all the masters while collecting each
    # glyphset. Glyph order is ignored.
    all_gsets = set()
    each_gset = []
    for master in master_list:
        master_path = master.attrib['filename']
        ufo_path = os.path.join(os.path.dirname(ds_path), master_path)
        gset = set(defcon.Font(ufo_path).keys())
        all_gsets.update(gset)
        each_gset.append(gset)

    master_indexes = []
    for i, gset in enumerate(each_gset):
        if gset != all_gsets:
            master_indexes.append(i)

    return master_indexes


def buildTempDesignSpace(dsPath):
    """'dsPath' is the path to the original designspace file.
       Returns a modified designspace file (for generating temp UFO masters)
       and a list of paths to the UFO masters (that will be used for generating
       each master OTF).
    """
    ds = ET.parse(dsPath).getroot()
    masterList = ds.find('sources').findall('source')

    # Delete any existing instances
    instances = ds.find('instances')
    if instances:
        ds.remove(instances)
    ds.append(XMLElement('instances'))
    instances = ds.find('instances')

    # Don't be wasteful, generate only the necessary temp masters
    temp_masters_to_generate = _determine_which_masters_to_generate(dsPath)

    # Populate the <instances> element with the information needed
    # for generating the temp master(s)
    master_paths = []
    for i, master in enumerate(masterList):
        masterPath = master.attrib['filename']
        if i not in temp_masters_to_generate:
            master_paths.append(masterPath)
            continue
        instance = XMLElement('instance')
        instance.append(master.find('location'))
        instance.append(XMLElement('kerning'))
        instance.append(XMLElement('info'))
        tempMasterPath = os.path.splitext(masterPath)[0] + kTempUFOExt
        master_paths.append(tempMasterPath)
        instance.attrib['filename'] = tempMasterPath
        ufo_path = os.path.join(os.path.dirname(dsPath), masterPath)
        ufo_info = defcon.Font(ufo_path).info
        instance.attrib['familyname'] = ufo_info.familyName
        instance.attrib['stylename'] = ufo_info.styleName
        instance.attrib['postscriptfontname'] = ufo_info.postscriptFontName
        instances.append(instance)
    tempDSPath = os.path.splitext(dsPath)[0] + kTempDSExt
    with open(tempDSPath, "w") as fp:
        fp.write(tounicode(xmlToString(ds)))
    return tempDSPath, master_paths


def testGlyphSetsCompatible(dsPath):
    # check if the masters all have the same set of glyphs
    glyphSetsDiffer = False
    dsDirPath = os.path.dirname(dsPath)
    rootET = ET.parse(dsPath)
    master_paths = []
    ds = rootET.getroot()
    sourceET = ds.find('sources')
    masterList = sourceET.findall('source')
    for et in masterList:
        master_paths.append(et.attrib['filename'])
    master_path = os.path.join(dsDirPath, master_paths[0])
    baseGO = defcon.Font(master_path).keys()
    for master_path in master_paths[1:]:
        master_path = os.path.join(dsDirPath, master_path)
        if baseGO != defcon.Font(master_path).keys():
            glyphSetsDiffer = True
            break
    return glyphSetsDiffer, master_paths


def parse_makeotf_options(mkot_opts):
    """Converts a comma-separated string into a space-separated string"""
    return ' '.join(mkot_opts.split(','))


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    if not args:
        args = ["-u"]

    mkot_options = ''

    if '-u' in args:
        print(__usage__)
        return
    if '-h' in args:
        print(__help__)
        return
    if '--mkot' in args:
        index = args.index('--mkot')
        args.pop(index)
        mkot_options = parse_makeotf_options(args.pop(index))

    (dsPath,) = args
    glyphSetsDiffer, master_paths = testGlyphSetsCompatible(dsPath)

    if glyphSetsDiffer:
        allowDecimalCoords = False
        logFile = tempfile.NamedTemporaryFile(delete=True).name
        tempDSPath, master_paths = buildTempDesignSpace(dsPath)
        print("Generating temp UFO master(s)...")
        ds_doc_reader = DesignSpaceDocumentReader(
            tempDSPath, ufoVersion=2, roundGeometry=(not allowDecimalCoords),
            verbose=False, logPath=logFile)
        ds_doc_reader.process(makeGlyphs=True, makeKerning=True, makeInfo=True)

    print("Building local otf's for master font paths...")
    curDir = os.getcwd()
    dsDir = os.path.dirname(dsPath)
    for master_path in master_paths:
        master_path = os.path.join(dsDir, master_path)
        masterDir = os.path.dirname(master_path)
        ufoName = os.path.basename(master_path)
        if ufoName.endswith(kTempUFOExt):
            otfName = ufoName[:-len(kTempUFOExt)]
            # copy the features.fea file from the original UFO master
            fea_file_path_from = os.path.join(
                master_path[:-len(kTempUFOExt)] + '.ufo', kFeaturesFile)
            fea_file_path_to = os.path.join(master_path, kFeaturesFile)
            if os.path.exists(fea_file_path_from):
                shutil.copyfile(fea_file_path_from, fea_file_path_to)
        else:
            otfName = os.path.splitext(ufoName)[0]
        otfName = otfName + ".otf"
        if masterDir:
            os.chdir(masterDir)
        cmd = "makeotf -nshw -f \"%s\" -o \"%s\" -r -nS %s 2>&1" % (
            ufoName, otfName, mkot_options)
        log = runShellCmd(cmd)
        if ("FATAL" in log) or ("Failed to build" in log):
            print(log)

        if "Built" not in str(log):
            print("Error building OTF font for", master_path, log)
            print("makeotf cmd was '%s' in %s." % (cmd, masterDir))
        else:
            print("Built OTF font for", master_path)
            compatibilizePaths(otfName)
            if ufoName.endswith(kTempUFOExt):
                shutil.rmtree(ufoName)
        os.chdir(curDir)

    if glyphSetsDiffer and os.path.exists(tempDSPath):
        os.remove(tempDSPath)


if __name__ == "__main__":
    main()
